#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Shape implementation using the
`shapely <http://toblerity.org/shapely/index.html>`_ library.
"""

import weakref
from copy import deepcopy
import logging

logger = logging.getLogger(__name__)

import shapely
from shapely.wkt import loads
from shapely.affinity import scale, translate
from shapely.geometry import Point, Polygon, MultiPolygon

import numpy as np
from numpy.random import uniform

from .geom_utils import conversion_magnitude
from .tools import indexable, pop_largest, _insert_area

# unit support

try:
    from .units import _unit_support
except ImportError:
    _unit_support = False

# opengl support

try:
    from .triangulate import triangulate, rnd_pts_in_tr
    _opengl_support = True
except ImportError:
    _opengl_support = False


__all__ = ["Area", "Shape"]


class Shape(Polygon):
    """
    Class containing the shape of the area where neurons will be distributed to
    form a network.

    Attributes
    ----------
    area : double
        Area of the shape in the :class:`Shape`'s
        :func:`Shape.unit` squared (:math:`\mu m^2`,
        :math:`mm^2`, :math:`cm^2`, :math:`dm^2` or :math:`m^2`).
    centroid : tuple of doubles
        Position of the center of mass of the current shape in `unit`.

    See also
    --------
    Parent class: :class:`shapely.geometry.Polygon`
    """

    @staticmethod
    def from_file(filename, min_x=None, max_x=None, unit='um', parent=None,
                  interpolate_curve=50, default_properties=None):
        '''
        Create a shape from a DXF, an SVG, or a WTK/WKB file.

        .. versionadded:: 0.3

        Parameters
        ----------
        filename : str
            Path to the file that should be loaded.
        min_x : float, optional (default: -5000.)
            Absolute horizontal position of the leftmost point in the
            environment in `unit` (default: 'um'). If None, no rescaling
            occurs.
        max_x : float, optional (default: 5000.)
            Absolute horizontal position of the rightmost point in the
            environment in `unit`. If None, no rescaling occurs.
        unit : string (default: 'um')
            Unit in the metric system among 'um' (:math:`\mu m`), 'mm', 'cm',
            'dm', 'm'.
        parent : :class:`nngt.Graph` object
            The parent which will become a :class:`nngt.SpatialGraph`.
        interpolate_curve : int, optional (default: 50)
            Number of points that should be used to interpolate a curve.
        default_properties : dict, optional (default: None)
            Default properties of the environment.
        '''
        from .shape_io import culture_from_file

        if _unit_support:
            from .units import Q_
            if isinstance(min_x, Q_):
                min_x = min_x.m_as(unit)
            if isinstance(max_x, Q_):
                max_x = max_x.m_as(unit)

        return culture_from_file(
                filename,  min_x=min_x, max_x=max_x, unit=unit, parent=parent,
                interpolate_curve=interpolate_curve,
                default_properties=default_properties)

    @staticmethod
    def from_polygon(polygon, min_x=None, max_x=None, unit='um',
                     parent=None, default_properties=None):
        '''
        Create a shape from a :class:`shapely.geometry.Polygon`.

        Parameters
        ----------
        polygon : :class:`shapely.geometry.Polygon`
            The initial polygon.
        min_x : float, optional (default: -5000.)
            Absolute horizontal position of the leftmost point in the
            environment in `unit` If None, no rescaling occurs.
        max_x : float, optional (default: 5000.)
            Absolute horizontal position of the rightmost point in the
            environment in `unit` If None, no rescaling occurs.
        unit : string (default: 'um')
            Unit in the metric system among 'um' (:math:`\mu m`), 'mm', 'cm',
            'dm', 'm'
        parent : :class:`nngt.Graph` object
            The parent which will become a :class:`nngt.SpatialGraph`.
        default_properties : dict, optional (default: None)
            Default properties of the environment.
        '''
        assert isinstance(polygon, Polygon), "`polygon` is not a Polygon " +\
            "but a {}.".format(polygon.__class__)

        if _unit_support:
            from .units import Q_
            if isinstance(min_x, Q_):
                min_x = min_x.m_as(unit)
            if isinstance(max_x, Q_):
                max_x = max_x.m_as(unit)

        obj    = None
        g_type = None

        # find the scaling factor
        if None not in (min_x, max_x):
            ext        = np.array(polygon.exterior.coords)
            leftmost   = np.min(ext[:, 0])
            rightmost  = np.max(ext[:, 0])
            scaling    = (max_x - min_x) / (rightmost - leftmost)
            obj        = scale(polygon, scaling, scaling)
        else:
            obj = Polygon(polygon)

        obj.__class__         = Shape
        obj._parent           = None
        obj._unit             = unit
        obj._geom_type        = g_type
        obj._return_quantity = False
        obj._areas = {
            "default_area": Area.from_shape(obj, name="default_area",
                                            properties=default_properties)
        }

        return obj

    @staticmethod
    def from_wkt(wtk, min_x=None, max_x=None, unit='um', parent=None,
                 default_properties=None):
        '''
        Create a shape from a WKT string.

        .. versionadded:: 0.2

        Parameters
        ----------
        wtk : str
            The WKT string.
        min_x : float, optional (default: -5000.)
            Absolute horizontal position of the leftmost point in the
            environment in `unit` If None, no rescaling occurs.
        max_x : float, optional (default: 5000.)
            Absolute horizontal position of the rightmost point in the
            environment in `unit` If None, no rescaling occurs.
        unit : string (default: 'um')
            Unit in the metric system among 'um' (:math:`\mu m`), 'mm', 'cm',
            'dm', 'm'
        parent : :class:`nngt.Graph` object
            The parent which will become a :class:`nngt.SpatialGraph`.
        default_properties : dict, optional (default: None)
            Default properties of the environment.

        See also
        --------
        :func:`Shape.from_polygon` for details about the other arguments.
        '''
        if _unit_support:
            from .units import Q_
            if isinstance(min_x, Q_):
                min_x = min_x.m_as(unit)
            if isinstance(max_x, Q_):
                max_x = max_x.m_as(unit)
        p = loads(wtk)
        return Shape.from_polygon(
            p, min_x=min_x, max_x=max_x, unit=unit, parent=parent,
            default_properties=default_properties)

    @staticmethod
    def rectangle(height, width, centroid=(0., 0.), unit='um',
                  parent=None, default_properties=None):
        '''
        Generate a rectangle of given height, width and center of mass.

        Parameters
        ----------
        height : float
            Height of the rectangle in `unit`
        width : float
            Width of the rectangle in `unit`
        centroid : tuple of floats, optional (default: (0., 0.))
            Position of the rectangle's center of mass in `unit`
        unit : string (default: 'um')
            Unit in the metric system among 'um' (:math:`\mu m`), 'mm', 'cm',
            'dm', 'm'
        parent : :class:`nngt.Graph` or subclass, optional (default: None)
            The parent container.
        default_properties : dict, optional (default: None)
            Default properties of the environment.

        Returns
        -------
        shape : :class:`Shape`
            Rectangle shape.
        '''
        if _unit_support:
            from .units import Q_
            if isinstance(width, Q_):
                width = width.m_as(unit)
            if isinstance(height, Q_):
                height = height.m_as(unit)
            if isinstance(centroid, Q_):
                centroid = centroid.m_as(unit)
            elif isinstance(centroid, Point):
                centroid = list(centroid.coords)[0]
            elif isinstance(centroid[0], Q_):
                centroid = (centroid[0].m_as(unit), centroid[1].m_as(unit))
        half_w = 0.5 * width
        half_h = 0.5 * height
        centroid = np.array(centroid)
        points = [centroid + [half_w, half_h],
                  centroid + [half_w, -half_h],
                  centroid - [half_w, half_h],
                  centroid - [half_w, -half_h]]
        shape = Shape(points, unit=unit, parent=parent,
                      default_properties=default_properties)
        shape._geom_type = "Rectangle"
        return shape

    @staticmethod
    def disk(radius, centroid=(0.,0.), unit='um', parent=None,
             default_properties=None):
        '''
        Generate a disk of given radius and center (`centroid`).

        Parameters
        ----------
        radius : float
            Radius of the disk in `unit`
        centroid : tuple of floats, optional (default: (0., 0.))
            Position of the rectangle's center of mass in `unit`
        unit : string (default: 'um')
            Unit in the metric system among 'um' (:math:`\mu m`), 'mm', 'cm',
            'dm', 'm'
        parent : :class:`nngt.Graph` or subclass, optional (default: None)
            The parent container.
        default_properties : dict, optional (default: None)
            Default properties of the environment.

        Returns
        -------
        shape : :class:`Shape`
            Rectangle shape.
        '''
        if _unit_support:
            from .units import Q_
            if isinstance(radius, Q_):
                radius = radius.m_as(unit)
            if isinstance(centroid, Q_):
                centroid = centroid.m_as(unit)
            elif isinstance(centroid, Point):
                centroid = list(centroid.coords)[0]
            elif isinstance(centroid[0], Q_):
                centroid = (centroid[0].m_as(unit), centroid[1].m_as(unit))
        centroid = np.array(centroid)
        minx = centroid[0] - radius
        maxx = centroid[0] + radius
        disk = Shape.from_polygon(
            Point(centroid).buffer(radius), min_x=minx, max_x=maxx, unit=unit,
            parent=parent, default_properties=default_properties)
        disk._geom_type = "Disk"
        disk.radius = radius
        return disk

    @staticmethod
    def ellipse(radii, centroid=(0.,0.), unit='um', parent=None,
                default_properties=None):
        '''
        Generate a disk of given radius and center (`centroid`).

        Parameters
        ----------
        radii : tuple of floats
            Couple (rx, ry) containing the radii of the two axes in `unit`
        centroid : tuple of floats, optional (default: (0., 0.))
            Position of the rectangle's center of mass in `unit`
        unit : string (default: 'um')
            Unit in the metric system among 'um' (:math:`\mu m`), 'mm', 'cm',
            'dm', 'm'
        parent : :class:`nngt.Graph` or subclass, optional (default: None)
            The parent container.
        default_properties : dict, optional (default: None)
            Default properties of the environment.

        Returns
        -------
        shape : :class:`Shape`
            Rectangle shape.
        '''
        if _unit_support:
            from .units import Q_
            if isinstance(radii, Q_):
                radii = radii.m_as(unit)
            elif isinstance(radii[0], Q_):
                radii = (radii[0].m_as(unit), radii[1].m_as(unit))
            if isinstance(centroid, Q_):
                centroid = centroid.m_as(unit)
            elif isinstance(centroid, Point):
                centroid = list(centroid.coords)[0]
            elif isinstance(centroid[0], Q_):
                centroid = (centroid[0].m_as(unit), centroid[1].m_as(unit))
        centroid = np.array(centroid)
        rx, ry = radii
        minx = centroid[0] - rx
        maxx = centroid[0] + rx
        ellipse = Shape.from_polygon(
            scale(Point(centroid).buffer(1.), rx, ry), min_x=minx, max_x=maxx,
            unit=unit, parent=parent, default_properties=default_properties)
        ellipse._geom_type = "Ellipse"
        ellipse.radii = radii
        return ellipse

    def __init__(self, shell, holes=None, unit='um', parent=None,
                 default_properties=None):
        '''
        Initialize the :class:`Shape` object and the underlying
        :class:`shapely.geometry.Polygon`.

        Parameters
        ----------
        exterior : array-like object of shape (N, 2)
            List of points defining the external border of the shape.
        interiors : array-like, optional (default: None)
            List of array-like objects of shape (M, 2), defining empty regions
            inside the shape.
        unit : string (default: 'um')
            Unit in the metric system among 'um' (:math:`\mu m`), 'mm', 'cm',
            'dm', 'm'.
        parent : :class:`nngt.Graph` or subclass
            The graph which is associated to this Shape.
        default_properties : dict, optional (default: None)
            Default properties of the environment.
        '''
        if _unit_support:
            from .units import Q_
            if isinstance(shell, Q_):
                shell = shell.m_as(unit)
            else:
                try:
                    if isinstance(shell[0], Q_):
                        shell = [q.m_as(unit) for q in shell]
                except:
                    pass
            if holes is not None:
                for i, h in enumerate(holes):
                    if isinstance(h, Q_):
                        h = h.m_as(unit)
                    else:
                        try:
                            if isinstance(h[0], Q_):
                                holes[i] = [q.m_as(unit) for q in h]
                        except:
                            pass

        self._return_quantity = False

        self._parent    = weakref.proxy(parent) if parent is not None else None
        self._unit      = unit
        self._geom_type = 'Polygon'

        # create the default area
        tmp = Polygon(shell, holes=holes)
        self._areas     = {
            "default_area": Area.from_shape(
                tmp, name="default_area", properties=default_properties,
                unit=unit)
        }

        super(Shape, self).__init__(shell, holes=holes)

    def copy(self):
        '''
        Create a copy of the current Shape.
        '''
        copy = Shape.from_polygon(self)

        copy._return_quantity = self._return_quantity

        copy._parent    = None
        copy._unit      = self._unit
        copy._geom_type = self._geom_type

        if self._areas:
            copy._areas     = {
                key: val.copy() for key, val in self._areas.items()
            }

        return copy

    @property
    def parent(self):
        ''' Return the parent of the :class:`Shape`. '''
        return self._parent

    @property
    def unit(self):
        '''
        Return the unit for the :class:`Shape` coordinates.
        '''
        return self._unit

    @property
    def areas(self):
        '''
        Returns the dictionary containing the Shape's areas.
        '''
        return deepcopy(self._areas)

    @property
    def default_areas(self):
        '''
        Returns the dictionary containing only the default areas.

        .. versionadded:: 0.4
        '''
        areas = {
            k: deepcopy(v) for k,v in self._areas.items()
            if k.find("default_area") == 0
        }
        return areas

    @property
    def non_default_areas(self):
        '''
        Returns the dictionary containing all Shape's areas except the
        default ones.

        .. versionadded:: 0.4
        '''
        areas = {
            k: deepcopy(v) for k,v in self._areas.items()
            if k.find("default_area") != 0
        }
        return areas

    @property
    def return_quantity(self):
        '''
        Whether `seed_neurons` returns positions with units by default.

        .. versionadded:: 0.5
        '''
        return self._return_quantity

    def add_area(self, area, height=None, name=None, properties=None,
                 override=False):
        '''
        Add a new area to the :class:`Shape`.
        If the new area has a part that is outside the main :class:`Shape`,
        it will be cut and only the intersection between the area and the
        container will be kept.

        Parameters
        ----------
        area : :class:`Area` or :class:`Shape`, or :class:`shapely.Polygon`.
            Delimitation of the area. Only the intersection between the parent
            :class:`Shape` and this new area will be kept.
        name : str, optional, default ("areaX" where X is the number of areas)
            Name of the area, under which it can be retrieved using the
            :func:`Shape.area` property of the :class:`Shape` object.
        properties : dict, optional (default: None)
            Properties of the area. If `area` is a :class:`Area`, then this is
            not necessary.
        override : bool, optional (default: False)
            If True, the new area will be made over existing areas that will
            be reduced in consequence.
        '''
        # check that area and self overlap
        assert self.overlaps(area) or self.contains(area), "`area` must be " +\
            "contained or at least overlap with the current shape."
        # check units
        if _unit_support:
            from .units import Q_
            if isinstance(height, Q_):
                height = height.m_as(self.unit)
        # check whether this area intersects with existing areas other than
        # the default area.
        intersection = self.intersection(area)
        if not override:
            for key, other_area in self._areas.items():
                if key.find("default_area") == -1:
                    assert not intersection.overlaps(other_area), \
                        "Different areas of a given Shape should not overlap."
        else:
            delete = []
            for key, other_area in self.non_default_areas.items():
                if other_area.overlaps(area) or other_area.contains(area):
                    new_existing = other_area.difference(area)
                    if new_existing.empty():
                        delete.append(key)
                    else:
                        _insert_area(self, key, new_existing,
                                     other_area.height, other_area.properties)
            for key in delete:
                del self._areas[key]

        # check properties
        if name is None:
            if isinstance(area, Area):
                name = area.name
            else:
                name = "area{}".format(len(self._areas))
        if height is None:
            if isinstance(area, Area):
                height = area.height
            else:
                height = self.areas["default_area"].height
        if properties is None:
            if isinstance(area, Area):
                properties = area.properties
            else:
                properties = {}
        # update the default area
        default_area = self._areas["default_area"]
        new_default = default_area.difference(intersection)

        # check that we do not add an area containing default
        if not new_default.is_empty:
            _insert_area(self, "default_area", new_default,
                         default_area.height, default_area.properties)
            # create the area
            _insert_area(self, name, intersection, height, properties)

    def add_hole(self, hole):
        '''
        Make a hole in the shape.

        .. versionadded:: 0.4
        '''
        areas = self.areas.copy()
        new_shape  = Shape.from_polygon(
            self.difference(hole), unit=self.unit, parent=self.parent,
            default_properties=areas["default_area"].properties)

        self._geom             = new_shape._geom
        new_shape._other_owned = True

        for name, area in areas.items():
            if name.find("default_area") != 0:
                _insert_area(self, name, area.difference(hole),
                             area.height, area.properties)

    def random_obstacles(self, n, form, params=None, heights=None,
                         properties=None, etching=0, on_area=None):
        '''
        Place random obstacles inside the shape.

        .. versionadded:: 0.4

        Parameters
        ----------
        n : int or float
            Number of obstacles if `n` is an :obj:`int`, otherwise represents
            the fraction of the shape's bounding box that should be occupied by
             the obstacles' bounding boxes.
        form : str or Shape
            Form of the obstacles, among "disk", "ellipse", "rectangle", or a
            custom shape.
        params : dict, optional (default: None)
            Dictionnary containing the instructions to build a predefined form
            ("disk", "ellipse", "rectangle"). See their creation methods for
            details. Leave `None` when using a custom shape.
        heights : float or list, optional (default: None)
            Heights of the obstacles. If None, the obstacle will considered as
            a "hole" in the structure, i.e. an uncrossable obstacle.
        properties : dict or list, optional (default: None)
            Properties of the obstacles if they constitue areas (only used if
            `heights` is not None). If not provided and `heights` is not None,
            will default to the "default_area" properties.
        etching : float, optional (default: 0)
            Etching of the obstacles' corners (rounded corners). Valid only
            for
        '''
        form_center = None

        if heights is not None:
            if _unit_support:
                from .units import Q_
                if isinstance(heights, Q_):
                    heights = heights.m_as(self.unit)
                elif indexable(heights):
                    if isinstance(heights[0], Q_):
                        heights = [h.m_as(self.unit) for h in heights]

        # check n
        if not isinstance(n, np.integer):
            assert n <= 1, "Filling fraction (floating point `n`) must be "  +\
                           "smaller or equal to 1."

        # check form
        if form == "disk":
            form = self.disk(**params)
        elif form == "ellipse":
            form = self.ellipse(**params)
        elif form == "rectangle":
            form = self.rectangle(**params)
        elif not isinstance(form, (Polygon, MultiPolygon, Shape, Area)):
            raise RuntimeError("Invalid form: '{}'.".format(form))

        # get form center and center on (0, 0)
        xmin, ymin, xmax, ymax = form.bounds
        form_center            = (0.5*(xmax + xmin), 0.5*(ymax + ymin))
        form_width             = xmax - xmin
        form_height            = ymax - ymin
        form_bbox_area         = float((xmax - xmin)*(ymax - ymin))

        # get shape width and height
        xmin, ymin, xmax, ymax = self.bounds
        width                  = xmax - xmin
        height                 = ymax - ymin

        if not np.allclose(form_center, (0, 0)):
            form = translate(form, -form_center[0], -form_center[1])

        # create points where obstacles can be located
        locations = []
        on_width  = int(np.rint(width / form_width))
        on_height = int(np.rint(height / form_height))
        x_offset  = 0.5*(width - on_width*form_width)
        y_offset  = 0.5*(height - on_height*form_height)

        for i in range(on_width):
            for j in range(on_height):
                x = xmin + x_offset + i*form_width
                y = ymin + y_offset + j*form_height
                locations.append((x, y))

        # get elected locations
        if not isinstance(n, np.integer):
            n = int(np.rint(len(locations) * n))

        indices   = list(range(len(locations)))
        indices   = np.random.choice(indices, n, replace=False)
        locations = [locations[i] for i in indices]

        # check heights
        same_prop = []
        if heights is not None:
            try:
                if len(heights) != n:
                    raise RuntimeError("One `height` entry per obstacle is "
                                       "required; expected "
                                       "{} but got {}".format(n, len(heights)))
                same_prop.append(np.allclose(heights, heights[0]))
            except TypeError:
                same_prop.append(True)
                heights     = [heights for _ in range(n)]

        # check properties
        if isinstance(properties, dict):
            properties = (properties for _ in range(n))
            same_prop.append(True)
        elif properties is not None:
            assert len(properties) == n, \
                "One `properties` entry per obstacle is  required; " +\
                "expected {} but got {}".format(n, len(properties))
            same_prop.append(True)
            for dic in properties:
                same_prop[-1] *= (dic == properties[0])
        else:
            same_prop.append(True)
            properties = (
                self.areas["default_area"].properties.copy() for _ in range(n)
            )

        # make names
        num_obstacles = 0
        for name in self.areas:
            if name.find("obstacle_") == 0:
                num_obstacles += 1

        names = ["obstacle_{}".format(num_obstacles + i) for i in range(n)]

        # create the obstacles
        if heights is None:
            new_form = Polygon()
            for loc in locations:
                new_form = new_form.union(translate(form, loc[0], loc[1]))
            if etching > 0:
                new_form = new_form.buffer(-etching, cap_style=3)
                new_form = new_form.buffer(etching)
            self.add_hole(new_form)
        else:
            if np.all(same_prop):
                # potentially contiguous areas
                new_form = Polygon()
                h        = next(iter(heights))
                prop     = next(iter(properties))
                for loc in locations:
                    new_form = new_form.union(translate(form, loc[0], loc[1]))
                if etching > 0:
                    new_form = new_form.buffer(-etching, cap_style=3)
                    new_form = new_form.buffer(etching)
                if self.overlaps(new_form) or self.contains(new_form):
                    self.add_area(new_form, height=h, name="obstacle",
                                  properties=prop, override=True)
            else:
                # many separate areas
                prop = (locations, heights, names, properties)
                for loc, h, name, p in zip(*prop):
                    new_form = translate(form, loc[0], loc[1])
                    if etching > 0:
                        new_form = new_form.buffer(-etching, cap_style=3)
                        new_form = new_form.buffer(etching)
                    if h is None:
                        self.add_hole(new_form)
                    elif self.overlaps(new_form) or self.contains(new_form):
                            self.add_area(new_form, height=h, name=name,
                                          properties=p, override=True)

    def set_parent(self, parent):
        ''' Set the parent :class:`nngt.Graph`. '''
        self._parent = weakref.proxy(parent) if parent is not None else None

    def set_return_units(self, b):
        '''
        Set the default behavior for positions returned by `seed_neurons`.
        If `True`, then the positions returned are quantities with units (from
        the `pint` library), otherwise they are simply numpy arrays.

        .. versionadded:: 0.5

        Note
        ----
        `set_return_units(True)` requires `pint` to be installed on the system,
        otherwise an error will be raised.
        '''
        if b and not _unit_support:
            raise RuntimeError("Cannot set 'return_quantity' to True as "
                               "`pint` is not installed.")
        self._return_quantity = b
        for area in self.areas.values():
            area._return_quantity = b

    def seed_neurons(self, neurons=None, container=None, on_area=None,
                     xmin=None, xmax=None, ymin=None, ymax=None, soma_radius=0,
                     unit=None, return_quantity=None):
        '''
        Return the positions of the neurons inside the
        :class:`Shape`.

        Parameters
        ----------
        neurons : int, optional (default: None)
            Number of neurons to seed. This argument is considered only if the
            :class:`Shape` has no `parent`, otherwise, a position is generated
            for each neuron in `parent`.
        container : :class:`Shape`, optional (default: None)
            Subshape acting like a mask, in which the neurons must be
            contained. The resulting area where the neurons are generated is
            the :func:`~shapely.Shape.intersection` between of the current
            shape and the `container`.
        on_area : str or list, optional (default: None)
            Area(s) where the seeded neurons should be.
        xmin : double, optional (default: lowest abscissa of the Shape)
            Limit the area where neurons will be seeded to the region on the
            right of `xmin`.
        xmax : double, optional (default: highest abscissa of the Shape)
            Limit the area where neurons will be seeded to the region on the
            left of `xmax`.
        ymin : double, optional (default: lowest ordinate of the Shape)
            Limit the area where neurons will be seeded to the region on the
            upper side of `ymin`.
        ymax : double, optional (default: highest ordinate of the Shape)
            Limit the area where neurons will be seeded to the region on the
            lower side of `ymax`.
        unit : string (default: None)
            Unit in which the positions of the neurons will be returned, among
            'um', 'mm', 'cm', 'dm', 'm'.
        return_quantity : bool, optional (default: False)
            Whether the positions should be returned as ``pint.Quantity``
            objects (requires Pint).

        .. versionchanged:: 0.5
            Accepts `pint` units and `return_quantity` argument.

        Note
        ----
        If both `container` and `on_area` are provided, the intersection of
        the two is used.

        Returns
        -------
        positions : array of double with shape (N, 2) or `pint.Quantity` if
                    `return_quantity` is `True`.
        '''
        return_quantity = (self._return_quantity
                           if return_quantity is None else return_quantity)

        if return_quantity:
            unit = self._unit if unit is None else unit
            if not _unit_support:
                raise RuntimeError("`return_quantity` requested but Pint is "
                                   "not available. Please install it first.")
        if _unit_support:
            from .units import Q_
            if isinstance(xmin, Q_):
                xmin = xmin.m_as(unit)
            if isinstance(xmax, Q_):
                xmax = xmax.m_as(unit)
            if isinstance(ymin, Q_):
                ymin = ymin.m_as(unit)
            if isinstance(ymax, Q_):
                ymax = ymax.m_as(unit)
            if isinstance(soma_radius, Q_):
                soma_radius = soma_radius.m_as(unit)

        positions = None
        if neurons is None and self._parent is not None:
            neurons = self._parent.node_nb()
        if neurons is None:
            raise ValueError("`neurons` cannot be None if `parent` is None.")
        if on_area is not None:
            if not hasattr(on_area, '__iter__'):
                on_area = [on_area]

        min_x, min_y, max_x, max_y = self.bounds

        custom_shape = (container is not None)
        if container is None and on_area is None:
            # set min/max
            if xmin is None:
                xmin = -np.inf
            if ymin is None:
                ymin = -np.inf
            if xmax is None:
                xmax = np.inf
            if ymax is None:
                ymax = np.inf
            min_x = max(xmin, min_x)  # smaller that Shape max x
            assert min_x <= self.bounds[2], "`min_x` must be inside Shape."
            min_y = max(ymin, min_y)  # smaller that Shape max y
            assert min_y <= self.bounds[3], "`min_y` must be inside Shape."
            max_x = min(xmax, max_x)  # larger that Shape min x
            assert max_x >= self.bounds[0], "`max_x` must be inside Shape."
            max_y = min(ymax, max_y)  # larger that Shape min y
            assert max_y >= self.bounds[1], "`max_y` must be inside Shape."
            # remaining tests
            if self._geom_type == "Rectangle":
                xx = uniform(
                    min_x + soma_radius, max_x - soma_radius, size=neurons)
                yy = uniform(
                    min_y + soma_radius, max_y - soma_radius, size=neurons)
                positions = np.vstack((xx, yy)).T
            elif (self._geom_type == "Disk"
                  and (xmin, ymin, xmax, ymax) == self.bounds):
                theta = uniform(0, 2*np.pi, size=neurons)
                # take some precaution to stay inside the shape
                r = (self.radius - soma_radius) *\
                    np.sqrt(uniform(0, 0.99, size=neurons))
                positions = np.vstack(
                    (r*np.cos(theta) + self.centroid[0],
                     r*np.sin(theta) + self.centroid[1])).T
            else:
                custom_shape = True
                container = Polygon([(min_x, min_y), (min_x, max_y),
                                     (max_x, max_y), (max_x, min_y)])
        elif on_area is not None:
            custom_shape = True
            area_shape   = Polygon()
            for area in on_area:
                area_shape = area_shape.union(self._areas[area])
            if container is not None:
                container = container.intersection(area_shape)
            else:
                container = area_shape
            assert container.area > 0, "`container` and `on_area` have " +\
                                       "empty intersection."

        # enter here only if Polygon or `container` is not None
        if custom_shape:
            seed_area   = self.intersection(container)
            area_buffer = seed_area.buffer(-soma_radius)
            if not area_buffer.is_empty:
                seed_area = area_buffer
            assert not seed_area.is_empty, "Empty area for seeding, check " +\
                "your `container` and min/max values."

            if not isinstance(seed_area, (Polygon, MultiPolygon)):
                raise ValueError("Invalid boundary value for seed region; "
                                 "check that the min/max values you requested "
                                 "are inside the shape.")

            if _opengl_support:
                triangles = []

                if isinstance(seed_area, MultiPolygon):
                    for g in seed_area.geoms:
                        triangles.extend((Polygon(v) for v in triangulate(g)))
                else:
                    triangles = [Polygon(v) for v in triangulate(seed_area)]

                positions = rnd_pts_in_tr(triangles, neurons)
            else:
                logger.warning("Random point generation can be very slow "
                               "without advanced triangulation methods. "
                               "Please install PyOpenGL for faster seeding "
                               "inside complex shapes.")
                points = []
                p = Point()
                while len(points) < neurons:
                    new_x = uniform(min_x, max_x, neurons-len(points))
                    new_y = uniform(min_y, max_y, neurons-len(points))
                    for x, y in zip(new_x, new_y):
                        p.coords = (x, y)
                        if seed_area.contains(p):
                            points.append((x, y))
                positions = np.array(points)

        if unit is not None and unit != self._unit:
            positions *= conversion_magnitude(unit, self._unit)

        if _unit_support and return_quantity:
            from .units import Q_
            return positions * Q_("um" if unit is None else unit)

        return positions

    def contains_neurons(self, positions):
        '''
        Check whether the neurons are contained in the shape.

        .. versionadded:: 0.4

        Parameters
        ----------
        positions : point or 2D-array of shape (N, 2)

        Returns
        -------
        contained : bool or 1D boolean array of length N
            True if the neuron is contained, False otherwise.
        '''
        if _unit_support:
            from .units import Q_
            if isinstance(positions, Q_):
                positions = positions.m_as(self._unit)
            elif len(positions):
                if isinstance(positions[0], Q_):
                    positions = np.array(
                        [q.m_as(self._unit) for q in positions])
                elif isinstance(positions[0][0], Q_):
                    positions = np.array(
                        [[q.m_as(self._unit) for q in row]
                         for row in positions])

        if np.shape(positions) == (len(positions), 2):
            contained = []
            for pos in positions:
                contained.append(self.contains(Point(*pos)))
            return np.array(contained, dtype=bool)
        else:
            return self.contains(Point(*positions))


class Area(Shape):
    """
    Specialized :class:`Shape` that stores additional properties regarding the
    interactions with the neurons.

    Each Area is characteristic of a given substrate and height. These two
    properties are homogeneous over the whole area, meaning that the neurons
    interact in the same manner with an Area reagardless of their position
    inside.

    The substrate is described through its modulation of the neuronal
    properties compared to their default behavior.
    Thus, a given area will modulate the speed, wall affinity, etc, of the
    growth cones that are growing above it.
    """

    @classmethod
    def from_shape(cls, shape, height=0., name="area", properties=None,
                   unit='um', min_x=None, max_x=None):
        '''
        Create an :class:`Area` from a :class:`Shape` object.

        Parameters
        ----------
        shape : :class:`Shape`
            Shape that should be converted to an Area.

        Returns
        -------
        :class:`Area` object.
        '''
        if _unit_support:
            from .units import Q_
            if isinstance(height, Q_):
                height = height.m_as(unit)
            if isinstance(min_x, Q_):
                min_x = min_x.m_as(unit)
            if isinstance(max_x, Q_):
                max_x = max_x.m_as(unit)

        obj    = None
        g_type = None
        if isinstance(shape, MultiPolygon):
            g_type = "MultiPolygon"
        elif isinstance(shape, (Polygon, Shape, Area)):
            g_type = "Polygon"
        else:
            raise TypeError("Expected a Polygon or MultiPolygon object.")
        # find the scaling factor
        scaling = 1.
        if None not in (min_x, max_x):
            ext        = np.array(shape.exterior.coords)
            leftmost   = np.min(ext[:, 0])
            rightmost  = np.max(ext[:, 0])
            scaling    = (max_x - min_x) / (rightmost - leftmost)
            obj        = scale(shape, scaling, scaling)
        else:
            if g_type == "Polygon":
                obj    = Polygon(shape)
            else:
                obj    = MultiPolygon(shape)

        obj.__class__        = cls
        obj._parent          = None
        obj._unit            = unit
        obj._geom_type       = g_type
        obj.__class__        = Area
        obj._areas           = None
        obj.height           = height
        obj.name             = name
        obj._prop            = _PDict(
            {} if properties is None else deepcopy(properties))
        obj._return_quantity = False

        return obj

    def __init__(self, shell, holes=None, unit='um', height=0.,
                 name="area", properties=None):
        '''
        Initialize the :class:`Shape` object and the underlying
        :class:`shapely.geometry.Polygon`.

        Parameters
        ----------
        shell : array-like object of shape (N, 2)
            List of points defining the external border of the shape.
        holes : array-like, optional (default: None)
            List of array-like objects of shape (M, 2), defining empty regions
            inside the shape.
        unit : string (default: 'um')
            Unit in the metric system among 'um' (:math:`\mu m`), 'mm', 'cm',
            'dm', 'm'.
        height : float, optional (default: 0.)
            Height of the area.
        name : str, optional (default: "area")
            The name of the area.
        properties : dict, optional (default: default neuronal properties)
            Dictionary containing the list of the neuronal properties that
            are modified by the substrate. Since this describes how the default
            property is modulated, all values must be positive reals or NaN.
        '''
        if _unit_support:
            from .units import Q_
            if isinstance(height, Q_):
                height = height.m_as(unit)

        super(Area, self).__init__(shell, holes=holes, unit=unit, parent=None)

        self._areas = None
        self.height = height
        self.name   = name
        self._prop  = _PDict(
            {} if properties is None else deepcopy(properties))

    def __deepcopy__(self, *args, **kwargs):
        obj = Area.from_shape(
            self, height=self.height, name=self.name, properties=self._prop)
        return obj

    def copy(self):
        '''
        Create a copy of the current Area.
        '''
        return Area.from_shape(super().copy(), height=self.height,
                               name=self.name, properties=self._prop.todict())

    @property
    def areas(self):
        raise AttributeError("Areas do not have sub-Areas.")

    @property
    def properties(self):
        p = self._prop.copy()
        p["height"] = self.height
        return p

    def add_subshape(self, subshape, position, unit='um'):
        raise NotImplementedError("Areas cannot be modified.")


class _PDict(dict):
    """
    Modified dictionary storing the modulation of the properties of an
    :class:`Area`.
    """

    def __getitem__(self, key):
        '''
        Returns 1 if key is not present.
        '''
        return super(_PDict, self).__getitem__(key) if key in self else 1.

    def __setitem__(self, key, value):
        '''
        Check that the value is a positive real or NaN before setting it.
        '''
        if key != "substrate_affinity":
            assert value >= 0 or np.isnan(value), \
                "`{}` property must be a positive real or NaN.".format(key)
        else:
            assert isinstance(value, float), \
                "`substrate_affinity` must be real or NaN."
        super(_PDict, self).__setitem__(key, float(value))

    def todict(self):
        return {k: v for k, v in self.items()}



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_analysis.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

"""
Test the IO functions.
"""

import os
import unittest

import numpy as np
import pytest

import nngt
from base_test import TestBasis, XmlHandler, network_dir
from tools_testing import foreach_graph


# --------------------- #
# File creation/removal #
# --------------------- #

current_dir = os.path.dirname(os.path.abspath(__file__)) + '/'
error = 'Wrong {{val}} for {graph}.'

gfilename = current_dir + 'g.graph'


def teardown_function(function):
    ''' Cleanup the file '''
    if nngt.get_config("mpi"):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD.Clone()

        comm.Barrier()

    try:
        os.remove(gfilename)
    except:
        pass


# ---------- #
# Test class #
# ---------- #

class TestIO(TestBasis):

    '''
    Class testing saving and loading functions.
    '''

    @classmethod
    def tearDownClass(cls):
        for graphname in cls.graphs:
            try:
                os.remove(current_dir + graphname + '.el')
            except:
                pass
        try:
            for fmt in ("nn", "el", "gml"):
                os.remove(current_dir + 'test.' + fmt)
        except:
            pass
    
    @property
    def test_name(self):
        return "test_io"

    @unittest.skipIf(nngt.get_config('mpi'), 'Not checking for MPI')
    def gen_graph(self, graph_name):
        # check whether we are loading from file
        if "." in graph_name:
            with_nngt = nngt.get_config("backend") == "nngt"

            if "graphml" in graph_name and with_nngt:
                return None

            abspath = network_dir + graph_name
            di_instructions = self.parser.get_graph_options(graph_name)
            graph = nngt.Graph.from_file(abspath, **di_instructions,
                                         cleanup=True)
            graph.set_name(graph_name)
            return graph, None
        else:
            di_instructions = self.parser.get_graph_options(graph_name)
            graph = nngt.generate(di_instructions)
            graph.set_name(graph_name)
            graph.to_file(current_dir + graph_name + '.el')

            return graph, di_instructions

    @foreach_graph
    def test_identical(self, graph, instructions, **kwargs):
        '''
        Test that the generated graph and the one loaded from the saved file
        are indeed identical.
        '''
        err = error.format(graph=graph.name)
        if instructions is not None:  # working with generated graph
            # load graph
            h = nngt.Graph.from_file(current_dir + graph.name + '.el')
            attributes = h.edge_attributes

            # test properties
            self.assertTrue(h.node_nb() == graph.node_nb(),
                            err.format(val='node number'))
            self.assertTrue(h.edge_nb() == graph.edge_nb(),
                            err.format(val='edge number'))

            if graph.is_spatial():
                self.assertTrue(np.allclose(h.get_positions(),
                                            graph.get_positions()),
                                err.format(val='positions'))

            for attr, values in graph.edge_attributes.items():
                # different results probably because of rounding problems
                # note, here we are using the edge list format so edge order
                # is the same in both graphs
                new_val = h.get_edge_attributes(name=attr)
                allclose = np.allclose(new_val, values, 1e-4)
                if not allclose:
                    print("Error: expected")
                    print(values)
                    print("but got")
                    print(new_val)
                    print("max error is: {}".format(
                        np.max(np.abs(np.subtract(
                            new_val, values)))))
                self.assertTrue(allclose, err.format(val=attr))
        else:  # working with loaded graph
            nodes = self.get_expected_result(graph, "nodes")
            edges = self.get_expected_result(graph, "edges")
            directed = self.get_expected_result(graph, "directed")
            # check
            self.assertEqual(
                nodes, graph.node_nb(), err.format(val='node number'))
            self.assertEqual(
                edges, graph.edge_nb(), err.format(val='edge number'))
            self.assertTrue(directed == graph.is_directed(),
                            err.format(val='directedness'))

    @unittest.skipIf(nngt.get_config('mpi'), 'Not checking for MPI')
    def test_custom_attributes(self):
        '''
        Test that custom attributes are saved and loaded correctly
        '''
        num_nodes = 100
        avg_deg = 10

        g = nngt.Graph(nodes=num_nodes)
        g.new_edge_attribute("test_attr", "int")

        for i in range(num_nodes):
            targets = np.random.choice(num_nodes, size=avg_deg, replace=False)
            elist = np.zeros((len(targets), 2), dtype=int)
            elist[:, 0] = i
            elist[:, 1] = targets
            ids  = np.random.randint(0, avg_deg*num_nodes, len(targets))
            ids *= 2*np.random.randint(0, 2, len(targets)) - 1
            g.new_edges(elist, attributes={"test_attr": ids},
                        check_duplicates=False, check_self_loops=False,
                        check_existing=False)

        old_edges = g.edges_array

        for fmt in ("nn", "el", "gml"):
            g.to_file(current_dir + 'test.' + fmt)
            h = nngt.Graph.from_file(current_dir + 'test.' + fmt)

            # for neighbour list, we need to give the edge list to have
            # the edge attributes in the same order as the original graph
            allclose = np.allclose(g.get_edge_attributes(name="test_attr"),
                                   h.get_edge_attributes(edges=old_edges,
                                                         name="test_attr"))
            if not allclose:
                print("Results differed for '{}'.".format(g.name))
                print("using file 'test.{}'.".format(fmt))
                print(g.get_edge_attributes(name="test_attr"))
                print(h.get_edge_attributes(edges=old_edges, name="test_attr"))
                with open(current_dir + 'test.' + fmt, 'r') as f:
                    for line in f.readlines():
                        print(line.strip())

            self.assertTrue(allclose)


@pytest.mark.mpi_skip
def test_empty_out_degree():
    g = nngt.Graph(2)

    g.new_edge(0, 1)

    for fmt in ("neighbour", "edge_list"):
        nngt.save_to_file(g, gfilename, fmt=fmt)

        h = nngt.load_from_file(gfilename, fmt=fmt)

        assert np.array_equal(g.edges_array, h.edges_array)


@pytest.mark.mpi_skip
def test_str_attributes():
    g = nngt.Graph(2)

    g.new_edge(0, 1)

    g.new_edge_attribute("type", "string")
    g.set_edge_attribute("type", val='odd')

    g.new_node_attribute("rnd", "string")
    g.set_node_attribute("rnd", values=["s'adf", 'sd fr"'])

    for fmt in ("neighbour", "edge_list"):
        nngt.save_to_file(g, gfilename, fmt=fmt)

        h = nngt.load_from_file(gfilename, fmt=fmt)

        assert np.array_equal(g.edges_array, h.edges_array)

        assert np.array_equal(g.edge_attributes["type"],
                              h.edge_attributes["type"])

        assert np.array_equal(g.node_attributes["rnd"],
                              h.node_attributes["rnd"])


@pytest.mark.mpi_skip
def test_structure():
    # with a structure
    room1 = nngt.Group(25)
    room2 = nngt.Group(50)
    room3 = nngt.Group(40)
    room4 = nngt.Group(35)

    names = ["R1", "R2", "R3", "R4"]

    struct = nngt.Structure.from_groups((room1, room2, room3, room4), names)

    g = nngt.Graph(structure=struct)

    g.to_file(gfilename, fmt="edge_list")

    h = nngt.load_from_file(gfilename, fmt="edge_list")

    assert g.structure == h.structure

    # with a neuronal population
    g = nngt.Network.exc_and_inhib(100)

    g.to_file(gfilename, fmt="edge_list")

    h = nngt.load_from_file(gfilename, fmt="edge_list")

    assert g.population == h.population


@pytest.mark.mpi_skip
def test_node_attributes():
    num_nodes = 10
    g = nngt.generation.erdos_renyi(nodes=num_nodes, avg_deg=3, directed=False)

    g.new_node_attribute("size", "int", [2*(i+1) for i in range(num_nodes)])

    g.to_file(gfilename, fmt="edge_list")

    h = nngt.load_from_file(gfilename, fmt="edge_list")

    assert np.array_equal(g.node_attributes["size"], h.node_attributes["size"])


# ---------- #
# Test suite #
# ---------- #

suite = unittest.TestLoader().loadTestsFromTestCase(TestIO)

if __name__ == "__main__":
    if not nngt.get_config("mpi"):
        test_empty_out_degree()
        test_str_attributes()
        test_structure()
        test_node_attributes()
        unittest.main()



#-*- coding:utf-8 -*-
#
# lib/constants.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Constant values for NNGT """


__all__ = [
    "default_neuron",
    "default_synapse",
    "POS",
    "BWEIGHT",
    "DIST",
    "WEIGHT",
    "DELAY",
    "TYPE",
]


# ----- #
# Names #
# ----- #

POS = "position"
DIST = "distance"
WEIGHT = "weight"
BWEIGHT = "bweight"
DELAY = "delay"
TYPE = "type"


# ------------ #
# Basic values #
# ------------ #

default_neuron = "aeif_cond_alpha"
''' :class:`string`, the default NEST neuron model '''
default_synapse = "static_synapse"
''' :class:`string`, the default NEST synaptic model '''
default_delay = 1.
''' :class:`double`, the default synaptic delay in NEST '''



#-*- coding:utf-8 -*-
#
# generation/connectors.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Connector functions """

import numpy as np

import nngt
from nngt.generation import graph_connectivity as gc
from nngt.lib import is_iterable, nonstring_container
from nngt.lib.test_functions import deprecated
from nngt.lib.rng_tools import _generate_random


__all__ = [
    'connect_neural_groups',
    'connect_groups',
    'connect_neural_types',
    'connect_nodes'
]


# generator dictionary
_di_gen_edges = {
    "all_to_all": gc._all_to_all,
    "circular": gc._circular,
    "distance_rule": gc._distance_rule,
    "erdos_renyi": gc._erdos_renyi,
    "fixed_degree": gc._fixed_degree,
    "from_degree_list": gc._from_degree_list,
    "gaussian_degree": gc._gaussian_degree,
    "newman_watts": gc._newman_watts,
    "watts_strogatz": gc._watts_strogatz,
    "price_scale_free": gc._price_scale_free,
    "random_scale_free": gc._random_scale_free
}


_one_pop_models = {
    "circular", "newman_watts", "price_scale_free", "watts_strogatz",
}


def connect_nodes(network, sources, targets, graph_model, density=None,
                  edges=None, avg_deg=None, unit='um', weighted=True,
                  directed=True, multigraph=False, check_existing=True,
                  ignore_invalid=False, **kwargs):
    '''
    Function to connect nodes with a given graph model.

    .. versionchanged:: 2.0
        Added `check_existing` and `ignore_invalid` arguments.

    Parameters
    ----------
    network : :class:`Network` or :class:`SpatialNetwork`
        The network to connect.
    sources : list
        Ids of the source nodes.
    targets : list
        Ids of the target nodes.
    graph_model : string
        The name of the connectivity model (among "erdos_renyi",
        "random_scale_free", "price_scale_free", and "newman_watts").
    check_existing : bool, optional (default: True)
        Check whether some of the edges that will be added already exist in the
        graph.
    ignore_invalid : bool, optional (default: False)
        Ignore invalid edges: they are not added to the graph and are
        silently dropped. Unless this is set to true, an error is raised
        if an existing edge is re-generated.
    **kwargs : keyword arguments
        Specific model parameters. or edge attributes specifiers such as
        `weights` or `delays`.

    Note
    ----
    For graph generation methods which set the properties of a
    specific degree (e.g. :func:`~nngt.generation.gaussian_degree`), the
    nodes which have their property sets are the `sources`.
    '''
    if network.is_spatial() and 'positions' not in kwargs:
        kwargs['positions'] = network.get_positions().astype(np.float32).T
    if network.is_spatial() and 'shape' not in kwargs:
        kwargs['shape'] = network.shape

    if graph_model in _one_pop_models:
        assert np.array_equal(sources, targets), \
            "'" + graph_model + "' can only work on a single set of nodes."

    sources  = np.array(sources, dtype=np.uint)
    targets  = np.array(targets, dtype=np.uint)
    distance = []

    elist = _di_gen_edges[graph_model](
        sources, targets, density=density, edges=edges,
        avg_deg=avg_deg, weighted=weighted, directed=directed,
        multigraph=multigraph, distance=distance, **kwargs)

    # Attributes are not set by subfunctions
    attr = {}

    if 'weights' in kwargs:
        ww = kwargs['weights']

        if isinstance(ww, dict):
            attr['weight'] = _generate_random(len(elist), ww)
        else:
            attr['weight'] = ww

    if 'delays' in kwargs:
        dd = kwargs['delays']

        if isinstance(ww, dict):
            attr['delay'] = _generate_random(len(elist), dd)
        else:
            attr['delay'] = dd

    if network.is_spatial() and distance:
        attr['distance'] = distance

    # call only on root process (for mpi) unless using distributed backend
    if nngt.on_master_process() or nngt.get_config("backend") == "nngt":
        elist = network.new_edges(
            elist, attributes=attr, check_duplicates=False,
            check_self_loops=False, check_existing=check_existing,
            ignore_invalid=ignore_invalid)

    if not network._graph_type.endswith('_connect'):
        network._graph_type += "_nodes_connect"

    return elist


def connect_neural_types(network, source_type, target_type, graph_model,
                         density=None, edges=None, avg_deg=None, unit='um',
                         weighted=True, directed=True, multigraph=False,
                         check_existing=True, ignore_invalid=False, **kwargs):
    '''
    Function to connect excitatory and inhibitory population with a given graph
    model.

    .. versionchanged:: 2.0
        Added `check_existing` and `ignore_invalid` arguments.

    Parameters
    ----------
    network : :class:`Network` or :class:`SpatialNetwork`
        The network to connect.
    source_type : int or list
        The type of source neurons (``1`` for excitatory, ``-1`` for
        inhibitory neurons).
    target_type : int or list
        The type of target neurons.
    graph_model : string
        The name of the connectivity model (among "erdos_renyi",
        "random_scale_free", "price_scale_free", and "newman_watts").
    check_existing : bool, optional (default: True)
        Check whether some of the edges that will be added already exist in the
        graph.
    ignore_invalid : bool, optional (default: False)
        Ignore invalid edges: they are not added to the graph and are
        silently dropped. Unless this is set to true, an error is raised
        if an existing edge is re-generated.
    kwargs : keyword arguments
        Specific model parameters. or edge attributes specifiers such as
        `weights` or `delays`.

    Note
    ----
    For graph generation methods which set the properties of a
    specific degree (e.g. :func:`~nngt.generation.gaussian_degree`), the
    nodes which have their property sets are the `source_type`.
    '''
    assert network.is_network(), "This function requires a Network object."

    elist, source_ids, target_ids = None, [], []

    if network.is_spatial() and 'positions' not in kwargs:
        kwargs['positions'] = network.get_positions().astype(np.float32).T

    if network.is_spatial() and 'shape' not in kwargs:
        kwargs['shape'] = network.shape

    if not nonstring_container(source_type):
        source_type = [source_type]

    if not nonstring_container(target_type):
        target_type = [target_type]

    for group in network._population.values():
        if group.neuron_type in source_type:
            source_ids.extend(group.ids)

        if group.neuron_type in target_type:
            target_ids.extend(group.ids)

    source_ids = np.array(source_ids, dtype=np.uint)
    target_ids = np.array(target_ids, dtype=np.uint)

    elist = connect_nodes(
        network, source_ids, target_ids, graph_model, density=density,
        edges=edges, avg_deg=avg_deg, unit=unit, weighted=weighted,
        directed=directed, multigraph=multigraph,
        check_existing=check_existing, ignore_invalid=ignore_invalid,
        **kwargs)

    if not network._graph_type.endswith('_neural_type_connect'):
        network._graph_type += "_neural_type_connect"

    return elist


@deprecated("1.3.1", reason="the library is moving to more generic names",
            alternative="connect_groups", removal="3.0")
def connect_neural_groups(*args, **kwargs):
    ''' Deprecatd alias of :func:`connect_groups`. '''
    return connect_groups(*args, **kwargs)


def connect_groups(network, source_groups, target_groups, graph_model,
                   density=None, edges=None, avg_deg=None, unit='um',
                   weighted=True, directed=True, multigraph=False,
                   check_existing=True, ignore_invalid=False, **kwargs):
    '''
    Function to connect groups with a given graph model.

    .. versionchanged:: 2.0
        Added `check_existing` and `ignore_invalid` arguments.

    Parameters
    ----------
    network : :class:`Network` or :class:`SpatialNetwork`
        The network to connect.
    source_groups : str, :class:`NeuralGroup`, or iterable
        Names of the source groups (which contain the pre-synaptic neurons) or
        directly the group objects themselves.
    target_groups : str, :class:`NeuralGroup`, or iterable
        Names of the target groups (which contain the post-synaptic neurons) or
        directly the group objects themselves.
    graph_model : string
        The name of the connectivity model (among "erdos_renyi",
        "random_scale_free", "price_scale_free", and "newman_watts").
    check_existing : bool, optional (default: True)
        Check whether some of the edges that will be added already exist in the
        graph.
    ignore_invalid : bool, optional (default: False)
        Ignore invalid edges: they are not added to the graph and are
        silently dropped. Unless this is set to true, an error is raised
        if an existing edge is re-generated.
    kwargs : keyword arguments
        Specific model parameters. or edge attributes specifiers such as
        `weights` or `delays`.

    Note
    ----
    For graph generation methods which set the properties of a
    specific degree (e.g. :func:`~nngt.generation.gaussian_degree`), the
    groups which have their property sets are the `source_groups`.
    '''
    source_ids, target_ids = [], []

    if network.is_spatial():
        if 'positions' not in kwargs:
            kwargs['positions'] = network.get_positions().astype(np.float32).T
        if 'shape' not in kwargs:
            kwargs['shape'] = network.shape

    if isinstance(source_groups, str) or not is_iterable(source_groups):
        source_groups = [source_groups]
    if isinstance(target_groups, str) or not is_iterable(target_groups):
        target_groups = [target_groups]

    for s in source_groups:
        if isinstance(s, nngt.Structure):
            source_ids.extend(s.ids)
        elif isinstance(s, nngt.Group):
            source_ids.extend(s.ids)
        else:
            source_ids.extend(network.structure[s].ids)

    for t in target_groups:
        if isinstance(t, nngt.Structure):
            target_ids.extend(t.ids)
        elif isinstance(t, nngt.Group):
            target_ids.extend(t.ids)
        else:
            target_ids.extend(network.structure[t].ids)

    source_ids = np.array(source_ids, dtype=np.uint)
    target_ids = np.array(target_ids, dtype=np.uint)

    elist = connect_nodes(
        network, source_ids, target_ids, graph_model, density=density,
        edges=edges, avg_deg=avg_deg, unit=unit, weighted=weighted,
        directed=directed, multigraph=multigraph,
        check_existing=check_existing, ignore_invalid=ignore_invalid,
        **kwargs)

    if not network._graph_type.endswith('_neural_group_connect'):
        network._graph_type += "_neural_group_connect"

    return elist



#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2020  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Chord diagram
=============
"""

import matplotlib.pyplot as plt

import nngt

plt.rcParams["figure.facecolor"] = (0, 0, 0, 0)

nngt.seed(0)


# create a structured graph

room1 = nngt.Group(25)
room2 = nngt.Group(50)
room3 = nngt.Group(40)
room4 = nngt.Group(35)

names = ["R1", "R2", "R3", "R4"]

struct = nngt.Structure.from_groups((room1, room2, room3, room4), names)

g = nngt.Graph(structure=struct)

for room in struct:
    nngt.generation.connect_groups(g, room, room, "all_to_all")

nngt.generation.connect_groups(g, (room1, room2), struct, "erdos_renyi",
                               avg_deg=10, ignore_invalid=True)

nngt.generation.connect_groups(g, room3, room1, "erdos_renyi", avg_deg=20)

nngt.generation.connect_groups(g, room4, room3, "erdos_renyi", avg_deg=10)


# get the structure graph and plot

sg = g.get_structure_graph()

nngt.plot.chord_diagram(sg, names="name", sort="distance", fontcolor="grey",
                        use_gradient=True, show=True)



#-*- coding:utf-8 -*-
#
# geospatial/plot.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
import numpy as np

import cartopy.crs as ccrs

from ..plot import draw_network
from .countries import maps, convertors, country_points, country_codes


def draw_map(graph, node_names, geodata=None, geodata_names=None,
             points=None, show_points=False, linecolor=None, hue=None,
             proj=None, all_geodata=True, axis=None, show=False, **kwargs):
    '''
    Draw a network on a map.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Graph to plot.
    node_names : str
        Node attribute containing the nodes' names or A3 codes. This attribute
        will be used to place each node on the map. By default (if no `geodata`
        is provided), the world map is used and each node must therefore be
        associated to a country name or (better) an A3 ISO code.
    geodata : :class:`~geopandas.GeoDataFrame`, optional (default: world map)
        Optional dataframe containing the geospatial information.
        Predefined geodatas are "110m", "50m", and "10m" for world maps with
        respectively 110, 50, and 10 meter resolutions, or "adaptive" (default)
        for a world map with adaptive resolution depending on the country size.
    geodata_names : str, optional (default: "NAME_LONG" or "SU_A3")
        Column in `geodata` corresponding to the `node_names` (respectively
        for full country names or A3 codes).
    points : str, optional (default: capitals and representative points)
        Whether a precise point should be associated to each node.
        It can be either an entry in `geodata`, the "centroid" of each geometry
        entry, or a "representative" point.
        By default, if the world map is used, each country will be associated
        to its capital (contained in the module's ``cities`` object); if
        another `geodata` element is provided, it defaults to "representative".
    show_points : bool, optional (default: False)
        Wether the points should be displayed.
    linecolor : str, char, float or array, optional (default: current palette)
        Color of the map lines.
    esize : float, str, or array of floats, optional (default: 0.5)
        Width of the edges in percent of canvas length. Available string values
        are "betweenness" and "weight".
    ecolor : str, char, float or array, optional (default: "k")
        Edge color. If ecolor="groups", edges color will depend on the source
        and target groups, i.e. only edges from and toward same groups will
        have the same color.
    max_esize : float, optional (default: 5.)
        If a custom property is entered as `esize`, this normalizes the edge
        width between 0. and `max_esize`.
    threshold : float, optional (default: 0.5)
        Size under which edges are not plotted.
    proj : :mod:`cartopy.crs` object, optional (default: cartesian plane)
        Projection that will be used to draw the map.
    all_geodata : bool, optional (default: True)
        Whether all the data contained in `geodata` should be plotted, even if
        `graph` contains only a subset of it.
    axis : matplotlib axis, optional (default: a new axis)
        Axis that will be used to plot the graph.
    **kwargs : dict
        All possible arguments from :func:`~nngt.plot.draw_network`.
    '''
    names = graph.node_attributes[node_names]

    # check whether the names are full names or A3 codes
    is_a3_codes = True

    for n in names:
        if len(n) != 3:
            is_a3_codes = False
            break

    convert = None

    if geodata_names is None:
        if is_a3_codes:
            geodata_names = "SU_A3"
        else:
            geodata_names = "NAME_LONG"

    # set map
    world_map = True
    dataframe = None

    geodata = "adaptive" if geodata is None else geodata

    if isinstance(geodata, str):
        dataframe = maps[geodata]

        # update names
        if not is_a3_codes:
            names = [convertors.get(name, name) for name in names]
    else:
        world_map = False

    # projection
    if proj is None:
        proj = ccrs.PlateCarree()
    else:
        try:
            crs_proj4 = proj.proj4_init
            dataframe = dataframe.to_crs(crs_proj4)
        except:
            # PlateCarree
            pass

    if axis is None:
        fig = plt.figure()
        axis = plt.axes(projection=proj)

    # underlying map (optional)
    lw = kwargs.get("linewidth", 1)

    if all_geodata:
        dataframe.boundary.plot(ax=axis, color=linecolor, alpha=0.2, zorder=0,
                                linewidth=lw)

    # get existing elements
    mapping = None

    if is_a3_codes and isinstance(geodata, str):
        cc = country_codes[geodata]
        mapping = {n: cc[n] for n in names}
    else:
        mapping = {s[geodata_names]: i for i, s in dataframe.iterrows()}

    elements = [mapping[n] for n in names]

    if hue is None:
        dataframe.iloc[elements].boundary.plot(ax=axis, color=linecolor,
                                               zorder=1, linewidth=lw)
    else:
        if hue not in dataframe:
            dataframe[hue] = np.full(len(dataframe), np.NaN)

            dataframe.loc[elements, hue] = graph.node_attributes[hue]

        dataframe.iloc[elements].plot(column=hue, ax=axis, zorder=1,
                                      linewidth=lw)

    # get positions
    pos = []

    if points in dataframe:
        pos = [(p.xy[0][0], p.xy[1][0])
               for p in dataframe.iloc[elements, points]]
    elif points == "centroid":
        df = dataframe.loc[elements, "geometry"].centroid
        pos = [(p.xy[0][0], p.xy[1][0]) for p in df]
    elif points == "representative":
        df = dataframe.loc[elements, "geometry"].representative_point()
        pos = [(p.xy[0][0], p.xy[1][0]) for p in df]
    elif points is None and geodata in (None, "adaptive"):
        try:
            crs_proj4 = proj.proj4_init
            cpoints = country_points.to_crs(crs_proj4)
        except:
            # PlateCarree
            cpoints = country_points
        df = cpoints.loc[elements, "geometry"]
        pos = [(p.xy[0][0], p.xy[1][0]) for p in df]
    elif points is None:
        df  = dataframe.loc[elements, "geometry"].representative_point()
        pos = [(p.xy[0][0], p.xy[1][0]) for p in df]
    else:
        raise ValueError("Invalid value for `points`: {}".format(points))

    pos = np.array(pos)

    if "restrict_nodes" in kwargs:
        pos = pos[kwargs["restrict_nodes"]]

    rm_kw = [
        "show_environment", "positions", "axis", "tight", "fast", "spatial"
    ]

    for k in rm_kw:
        if k in kwargs:
            del kwargs[k]

    # make plot
    draw_network(graph, layout=pos, axis=axis, show_environment=False,
                 fast=True, tight=False, proj=proj, spatial=False, show=False,
                 **kwargs)

    # restore full map
    if all_geodata:
        axis.set_global()

    if show:
        plt.show()

    return axis



#-*- coding:utf-8 -*-
#
# geospatial/__init__.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
The geospatial module contains functions and objects that enable
straightforward network plots together with geospatial data.

It relies on `geopandas <geopandas.org>`_ and
`cartopy <https://scitools.org.uk/cartopy/docs/latest>`_ in the background.

See ":ref:`sphx_glr_gallery_graph_structure_plot_map.py`" for an example.
"""

from .countries import (maps, country_names, country_codes, convertors, cities,
                        codes_to_names, natural_earth)

from .plot import draw_map



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_attributes.py

"""
Test the validity of the plotting functions.
"""

import os

import numpy as np
import pytest

import nngt
import nngt.generation as ng
import nngt.plot as nplt


# absolute directory path

dirpath = os.path.abspath(os.path.dirname(__file__))


# tests
@pytest.mark.mpi_skip
def test_plot_config():
    ''' Test the default plot configuration '''
    nngt.set_config("color_lib", "seaborn")
    nngt.set_config("palette", "viridis")
    nngt.set_config("palette_discrete", "Set3")
    nngt.set_config("palette_continuous", "magma")


@pytest.mark.mpi_skip
def test_plot_prop():
    net = ng.erdos_renyi(nodes=100, avg_deg=10)
    net.set_weights(distribution="gaussian",
                    parameters={"avg": 5, "std": 0.5})
    net.new_node_attribute("attr", "int",
                           values=np.random.randint(-10, 20, 100))

    nplt.degree_distribution(net, ["in", "out", "total"], show=False)

    nplt.edge_attributes_distribution(net, "weight", show=False)

    nplt.node_attributes_distribution(net, "attr", show=False)

    if nngt.get_config("backend") != "nngt":
        nplt.betweenness_distribution(net, show=False)


@pytest.mark.mpi_skip
def test_draw_network_options():
    net = nngt.generation.erdos_renyi(nodes=100, avg_deg=10)

    net.set_weights(np.random.randint(0, 20, net.edge_nb()))

    nplt.draw_network(net, ncolor="in-degree", nsize=3, esize=2,
                      colorbar=True, show=False)

    if nngt.get_config("backend") != "nngt":
        nplt.draw_network(net, ncolor="betweenness", nsize="total-degree",
                          decimate_connections=3, curved_edges=True,
                          show=False)

    nplt.draw_network(net, ncolor="g", nshape='s', ecolor="b",
                      restrict_targets=[1, 2, 3], show=False)

    nplt.draw_network(net, restrict_nodes=list(range(10)), fast=True,
                      show=False)

    nplt.draw_network(net, restrict_targets=[4, 5, 6, 7, 8], show=False)

    nplt.draw_network(net, restrict_sources=[4, 5, 6, 7, 8], show=False)

    # plot on a single axis
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()

    nplt.draw_network(net, simple_nodes=True, ncolor="k",
                      decimate_connections=-1, axis=ax, show=False)

    nplt.draw_network(net, simple_nodes=True, ncolor="r", nsize=2,
                      restrict_nodes=list(range(10)), esize='weight',
                      ecolor="b", fast=True, axis=ax, show=False)


@pytest.mark.mpi_skip
def test_library_plot():
    ''' Check that plotting with the underlying backend library works '''
    pop = nngt.NeuralPop.exc_and_inhib(50)
    g   = ng.newman_watts(4, 0.2, population=pop)

    g.set_weights(np.random.uniform(1, 5, g.edge_nb()))

    nplt.library_draw(g, show=False)

    nplt.library_draw(g, ncolor="total-degree", ecolor="k", show=False)

    if nngt.get_config("backend") != "nngt":
        nplt.library_draw(g, ncolor="in-degree", ecolor="betweenness",
                          esize='weight', max_esize=5, show=False)

    nplt.library_draw(g, nshape="s", esize="weight", layout="random",
                      show=False)

    nplt.library_draw(g, nshape="s", esize="weight", layout="random",
                      show=False)

    nplt.library_draw(g, ncolor="in-degree", esize="weight", layout="circular",
                      show=False)

    edges = g.edges_array[::5]

    nplt.library_draw(g, ncolor="in-degree", esize="weight", layout="circular",
                      restrict_edges=edges, show=False)


@pytest.mark.mpi_skip
@pytest.mark.skipif(nngt.get_config("backend") == "nngt", reason="Lib needed.")
def test_hive_plot():
    g = nngt.load_from_file(dirpath + "/Networks/rat_brain.graphml",
                            attributes=["weight"], cleanup=True,
                            attributes_types={"weight": float})

    cc = nngt.analysis.local_clustering(g, weights="weight")

    g.new_node_attribute("cc", "double", values=cc)

    g.new_node_attribute("strength", "double",
                         values=g.get_degrees(weights="weight"))

    g.new_node_attribute(
        "SC", "double", values=nngt.analysis.subgraph_centrality(g))

    cc_bins = [0, 0.1, 0.25, 0.6]

    nplt.hive_plot(g, g.get_degrees(), axes="cc", axes_bins=cc_bins)

    nplt.hive_plot(g, "strength", axes="cc", axes_bins=cc_bins,
                   axes_units="rank")

    # set colormap and highlight nodes
    nodes = list(range(g.node_nb(), 3))

    nplt.hive_plot(g, "strength", axes="SC", axes_bins=4,
                   axes_colors="brg", highlight_nodes=nodes)

    # multiple radial axes and edge highlight
    rad_axes = ["cc", "strength", "SC"]
    edges    = g.get_edges(source_node=nodes)

    nplt.hive_plot(g, rad_axes, rad_axes,
                   nsize=g.get_degrees(), max_nsize=50, highlight_edges=edges)

    # check errors
    with pytest.raises(ValueError):
        nplt.hive_plot(g, [cc, "closeness", "strength"])
        nplt.hive_plot(g, 124)

    with pytest.raises(AssertionError):
        nplt.hive_plot(g, cc, axes=[1, 2, 3])
        nplt.hive_plot(g, cc, axes=1)
        nplt.hive_plot(g, cc, axes="groups")


@pytest.mark.mpi_skip
def test_plot_spatial_alpha():
    ''' Test positional layout and alpha parameters '''
    num_nodes = 4
    pos = [(1, 1), (0, -1), (-1, -1), (-1, 1)]

    g = nngt.SpatialGraph(num_nodes, positions=pos)
    g.new_edges([(0, 1), (0, 2), (1, 3), (3, 2)])

    nplt.draw_network(g, nsize=0.02, ealpha=1)

    nplt.draw_network(g, layout=[(y, x) for (x, y) in pos],
                      show_environment=False, nsize=0.02, nalpha=0.5)


if __name__ == "__main__":
    test_plot_config()
    test_plot_prop()
    test_draw_network_options()
    test_library_plot()
    test_hive_plot()
    test_plot_spatial_alpha()



#-*- coding:utf-8 -*-
#
# core/graph.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Graph class for graph generation and management """

import logging
import weakref

from collections import defaultdict
from copy import deepcopy

import numpy as np
import scipy.sparse as ssp

import nngt
import nngt.analysis as na

from nngt import save_to_file
from nngt.io.graph_loading import _load_from_file, _library_load
from nngt.io.io_helpers import _get_format
from nngt.io.graph_saving import _as_string
from nngt.lib import InvalidArgument, nonstring_container
from nngt.lib.connect_tools import _set_degree_type, _unique_rows
from nngt.lib.graph_helpers import _edge_prop, _get_matrices
from nngt.lib.logger import _log_message
from nngt.lib.test_functions import graph_tool_check, is_integer

from .connections import Connections


logger = logging.getLogger(__name__)


# ----- #
# Graph #
# ----- #

class Graph(nngt.core.GraphObject):

    """
    The basic graph class, which inherits from a library class such as
    :class:`graph_tool.Graph`, :class:`networkx.DiGraph`, or ``igraph.Graph``.

    The objects provides several functions to easily access some basic
    properties.
    """

    #-------------------------------------------------------------------------#
    # Class properties

    __num_graphs = 0
    __max_id = 0

    @classmethod
    def num_graphs(cls):
        ''' Returns the number of alive instances. '''
        return cls.__num_graphs

    @classmethod
    def from_library(cls, library_graph, name="ImportedGraph", weighted=True,
                     directed=True, **kwargs):
        '''
        Create a :class:`~nngt.Graph` by wrapping a graph object from one of
        the supported libraries.

        Parameters
        ----------
        library_graph : object
            Graph object from one of the supported libraries (graph-tool,
            igraph, networkx).
        name : str, optional (default: "ImportedGraph")
        **kwargs
            Other standard arguments (see :func:`~nngt.Graph.__init__`)
        '''
        graph = cls(name=name, weighted=False, **kwargs)

        graph._from_library_graph(library_graph, copy=False)

        return graph

    @classmethod
    def from_matrix(cls, matrix, weighted=True, directed=True, population=None,
                    shape=None, positions=None, name=None, **kwargs):
        '''
        Creates a :class:`~nngt.Graph` from a :mod:`scipy.sparse` matrix or
        a dense matrix.

        Parameters
        ----------
        matrix : :mod:`scipy.sparse` matrix or :class:`numpy.ndarray`
            Adjacency matrix.
        weighted : bool, optional (default: True)
            Whether the graph edges have weight properties.
        directed : bool, optional (default: True)
            Whether the graph is directed or undirected.
        population : :class:`~nngt.NeuralPop`
            Population to associate to the new :class:`~nngt.Network`.
        shape : :class:`~nngt.geometry.Shape`, optional (default: None)
            Shape to associate to the new :class:`~nngt.SpatialGraph`.
        positions : (N, 2) array
            Positions, in a 2D space, of the N neurons.
        name : str, optional
            Graph name.

        Returns
        -------
        :class:`~nngt.Graph`
        '''
        mshape = matrix.shape

        graph_name = "FromYMatrix_Z"

        nodes = max(mshape[0], mshape[1])

        if issubclass(matrix.__class__, ssp.spmatrix):
            graph_name = graph_name.replace('Y', 'Sparse')
            if not directed:
                if mshape[0] != mshape[1] or not (matrix.T != matrix).nnz == 0:
                    raise InvalidArgument('Incompatible `directed=False` '
                                          'option provided for non symmetric '
                                          'matrix.')

                matrix = ssp.tril(matrix, format=matrix.format)
        else:
            graph_name = graph_name.replace('Y', 'Dense')
            if not directed:
                if mshape[0] != mshape[1] or not (matrix.T == matrix).all():
                    raise InvalidArgument('Incompatible `directed=False` '
                                          'option provided for non symmetric '
                                          'matrix.')
                matrix = np.tril(matrix)

        edges = np.array(matrix.nonzero()).T

        graph_name = graph_name.replace("Z", str(cls.__num_graphs))

        # overwrite default name if necessary
        if name is not None:
            graph_name = name

        graph = cls(nodes, name=graph_name, weighted=weighted,
                    directed=directed, **kwargs)

        if population is not None:
            cls.make_network(graph, population)

        if shape is not None or positions is not None:
            cls.make_spatial(graph, shape, positions)

        weights = None

        if weighted:
            if issubclass(matrix.__class__, ssp.spmatrix):
                weights = np.array(matrix[edges[:, 0], edges[:, 1]])[0]
            else:
                weights = matrix[edges[:, 0], edges[:, 1]]

                if len(weights.shape) == 2:
                    weights = weights.A1

        attributes = {"weight": weights} if weighted else None

        graph.new_edges(edges, attributes, check_self_loops=False,
                        ignore_invalid=True)

        return graph

    @staticmethod
    def from_file(filename, fmt="auto", separator=" ", secondary=";",
                  attributes=None, attributes_types=None, notifier="@",
                  ignore="#", from_string=False, name=None,
                  directed=True, cleanup=False):
        '''
        Import a saved graph from a file.

        .. versionchanged :: 2.0
            Added optional `attributes_types` and `cleanup` arguments.

        Parameters
        ----------
        filename: str
            The path to the file.
        fmt : str, optional (default: deduced from filename)
            The format used to save the graph. Supported formats are:
            "neighbour" (neighbour list), "ssp" (scipy.sparse), "edge_list"
            (list of all the edges in the graph, one edge per line,
            represented by a ``source target``-pair), "gml" (gml format,
            default if `filename` ends with '.gml'), "graphml" (graphml format,
            default if `filename` ends with '.graphml' or '.xml'), "dot" (dot
            format, default if `filename` ends with '.dot'), "gt" (only
            when using `graph_tool <http://graph-tool.skewed.de/>`_ as library,
            detected if `filename` ends with '.gt').
        separator : str, optional (default " ")
            separator used to separate inputs in the case of custom formats
            (namely "neighbour" and "edge_list")
        secondary : str, optional (default: ";")
            Secondary separator used to separate attributes in the case of
            custom formats.
        attributes : list, optional (default: [])
            List of names for the attributes present in the file. If a
            `notifier` is present in the file, names will be deduced from it;
            otherwise the attributes will be numbered.
            For "edge_list", attributes may also be present as additional
            columns after the source and the target.
        attributes_types : dict, optional (default: str)
            Backup information if the type of the attributes is not specified
            in the file. Values must be callables (types or functions) that
            will take the argument value as a string input and convert it to
            the proper type.
        notifier : str, optional (default: "@")
            Symbol specifying the following as meaningfull information.
            Relevant information are formatted ``@info_name=info_value``, where
            ``info_name`` is in ("attributes", "directed", "name", "size") and
            associated ``info_value`` are of type (``list``, ``bool``, ``str``,
            ``int``).
            Additional notifiers are
            ``@type=SpatialGraph/Network/SpatialNetwork``, which must be
            followed by the relevant notifiers among ``@shape``,
            ``@population``, and ``@graph``.
        from_string : bool, optional (default: False)
            Load from a string instead of a file.
        ignore : str, optional (default: "#")
            Ignore lines starting with the `ignore` string.
        name : str, optional (default: from file information or 'LoadedGraph')
            The name of the graph.
        directed : bool, optional (default: from file information or True)
            Whether the graph is directed or not.
        cleanup : bool, optional (default: False)
           If true, removes nodes before the first one that appears in the
           edges and after the last one and renumber the nodes from 0.

        Returns
        -------
        graph : :class:`~nngt.Graph` or subclass
            Loaded graph.
        '''
        fmt = _get_format(fmt, filename)

        if fmt not in ("neighbour", "edge_list", "gml"):
            # only partial support for these formats, relying on backend
            libgraph = _library_load(filename, fmt)

            name = "LoadedGraph" if name is None else name

            graph = Graph.from_library(libgraph, name=name, directed=directed)

            return graph

        info, edges, nattr, eattr, struct, shape, pos = _load_from_file(
            filename=filename, fmt=fmt, separator=separator, ignore=ignore,
            secondary=secondary, attributes=attributes,
            attributes_types=attributes_types, notifier=notifier,
            cleanup=cleanup)

        # create the graph
        name = info.get("name", "LoadedGraph") if name is None else name

        graph = Graph(nodes=info["size"], name=name,
                      directed=info.get("directed", directed))

        # make the nodes attributes
        lst_attr, dtpes, lst_values = [], [], []

        if info["node_attributes"]:  # node attributes to add to the graph
            lst_attr   = info["node_attributes"]
            dtpes      = info["node_attr_types"]
            lst_values = [nattr[name] for name in info["node_attributes"]]

        for nattr, dtype, values in zip(lst_attr, dtpes, lst_values):
            graph.new_node_attribute(nattr, dtype, values=values)

        # make the edges and their attributes
        lst_attr, dtpes, lst_values = [], [], []

        if info["edge_attributes"]:  # edge attributes to add to the graph
            lst_attr   = info["edge_attributes"]
            dtpes      = info["edge_attr_types"]
            lst_values = [eattr[name] for name in info["edge_attributes"]]

        if len(edges):
            graph.new_edges(edges, check_duplicates=False,
                            check_self_loops=False, check_existing=False)

        for eattr, dtype, values in zip(lst_attr, dtpes, lst_values):
            graph.new_edge_attribute(eattr, dtype, values=values)

        if struct is not None:
            if isinstance(struct, nngt.NeuralPop):
                nngt.Network.make_network(graph, struct)
            else:
                graph.structure = struct

            struct._parent = weakref.ref(graph)

            for g in struct.values():
                g._struct = weakref.ref(struct)
                g._net    = weakref.ref(graph)

        if pos is not None or shape is not None:
            nngt.SpatialGraph.make_spatial(graph, shape=shape, positions=pos)

        return graph

    @staticmethod
    def make_spatial(graph, shape=None, positions=None, copy=False):
        '''
        Turn a :class:`~nngt.Graph` object into a :class:`~nngt.SpatialGraph`,
        or a :class:`~nngt.Network` into a :class:`~nngt.SpatialNetwork`.

        Parameters
        ----------
        graph : :class:`~nngt.Graph` or :class:`~nngt.SpatialGraph`
            Graph to convert.
        shape : :class:`~nngt.geometry.Shape`, optional (default: None)
            Shape to associate to the new :class:`~nngt.SpatialGraph`.
        positions : (N, 2) array
            Positions, in a 2D space, of the N neurons.
        copy : bool, optional (default: ``False``)
            Whether the operation should be made in-place on the object or if a
            new object should be returned.

        Notes
        -----
        In-place operation that directly converts the original graph if `copy`
        is ``False``, else returns the copied :class:`~nngt.Graph` turned into
        a :class:`~nngt.SpatialGraph`.
        The `shape` argument can be skipped if `positions` are given; in that
        case, the neurons will be embedded in a rectangle that contains them
        all.
        '''
        if copy:
            graph = graph.copy()

        if isinstance(graph, nngt.Network):
            graph.__class__ = nngt.SpatialNetwork
        else:
            graph.__class__ = nngt.SpatialGraph

        graph._init_spatial_properties(shape, positions)

        if copy:
            return graph

    @staticmethod
    def make_network(graph, neural_pop, copy=False, **kwargs):
        '''
        Turn a :class:`~nngt.Graph` object into a :class:`~nngt.Network`, or a
        :class:`~nngt.SpatialGraph` into a :class:`~nngt.SpatialNetwork`.

        Parameters
        ----------
        graph : :class:`~nngt.Graph` or :class:`~nngt.SpatialGraph`
            Graph to convert
        neural_pop : :class:`~nngt.NeuralPop`
            Population to associate to the new :class:`~nngt.Network`
        copy : bool, optional (default: ``False``)
            Whether the operation should be made in-place on the object or if a
            new object should be returned.

        Notes
        -----
        In-place operation that directly converts the original graph if `copy`
        is ``False``, else returns the copied :class:`~nngt.Graph` turned into
        a :class:`~nngt.Network`.
        '''
        if copy:
            graph = graph.copy()

        if isinstance(graph, nngt.SpatialGraph):
            graph.__class__ = nngt.SpatialNetwork
        else:
            graph.__class__ = nngt.Network

        # set delays to 1. or to provided value if they are not already set
        if "delays" not in kwargs and not hasattr(graph, '_d'):
            graph._d = {"distribution": "constant", "value": 1.}
        elif "delays" in kwargs and not hasattr(graph, '_d'):
            graph._d = kwargs["delays"]
        elif "delays" in kwargs:
            _log_message(logger, "WARNING",
                         'Graph already had delays set, ignoring new ones.')

        graph._init_bioproperties(neural_pop)

        if copy:
            return graph

    #-------------------------------------------------------------------------#
    # Constructor/destructor and properties

    def __new__(klass, *args, **kwargs):
        '''
        Create a new Graph object.
        '''
        has_pop = False
        is_sptl = False

        for arg in args:
            if isinstance(arg, nngt.geometry.Shape):
                is_sptl = True
            if isinstance(arg, nngt.NeuralPop):
                has_pop = True

        if "population" in kwargs:
            has_pop = True
        if "shape" in kwargs or "positions" in kwargs:
            is_sptl = True

        if is_sptl and has_pop:
            klass = nngt.SpatialNetwork
        elif is_sptl:
            klass = nngt.SpatialGraph
        elif has_pop:
            klass = nngt.Network

        return super().__new__(klass)
            

    def __init__(self, nodes=None, name="Graph", weighted=True, directed=True,
                 copy_graph=None, structure=None, **kwargs):
        '''
        Initialize Graph instance

        .. versionchanged:: 2.0
            Renamed `from_graph` to `copy_graph`.

        .. versionchanged:: 2.2
            Added `structure` argument.

        Parameters
        ----------
        nodes : int, optional (default: 0)
            Number of nodes in the graph.
        name : string, optional (default: "Graph")
            The name of this :class:`Graph` instance.
        weighted : bool, optional (default: True)
            Whether the graph edges have weight properties.
        directed : bool, optional (default: True)
            Whether the graph is directed or undirected.
        copy_graph : :class:`~nngt.Graph`, optional
            An optional :class:`~nngt.Graph` that will be copied.
        structure : :class:`~nngt.Structure`, optional (default: None)
            A structure dividing the graph into specific groups, which can
            be used to generate specific connectivities and visualise the
            connections in a more coarse-grained manner.
        kwargs : optional keywords arguments
            Optional arguments that can be passed to the graph, e.g. a dict
            containing information on the synaptic weights
            (``weights={"distribution": "constant", "value": 2.3}`` which is
            equivalent to ``weights=2.3``), the synaptic `delays`, or a
            ``type`` information.

        Note
        ----
        When using `copy_graph`, only the topological properties are
        copied (nodes, edges, and attributes), spatial and biological
        properties are ignored.
        To copy a graph exactly, use :func:`~nngt.Graph.copy`.

        Returns
        -------
        self : :class:`~nngt.Graph`
        '''
        self.__id = self.__class__.__max_id
        self._name = name
        self._graph_type = kwargs["type"] if "type" in kwargs else "custom"

        # check the structure
        if structure is not None:
            if nodes is None:
                nodes = structure.size
            else:
                assert nodes == structure.size, \
                    "`nodes` and `structure.size` must be the same."
        else:
            nodes = 0 if nodes is None else nodes

        self._struct = structure

        # Init the core.GraphObject
        super().__init__(nodes=nodes, copy_graph=copy_graph,
                         directed=directed, weighted=weighted)

        # take care of the weights and delays
        if copy_graph is None:
            if weighted:
                self.new_edge_attribute('weight', 'double')
                self._w = _edge_prop(kwargs.get("weights", None))
            if "delays" in kwargs:
                self.new_edge_attribute('delay', 'double')
                self._d = _edge_prop(kwargs.get("delays", None))
            if 'inh_weight_factor' in kwargs:
                self._iwf = kwargs['inh_weight_factor']
        else:
            self._w   = getattr(copy_graph, "_w", None)
            self._d   = getattr(copy_graph, "_d", None)
            self._iwf = getattr(copy_graph, "_iwf", None)

            self._eattr._num_values_set = \
                copy_graph._eattr._num_values_set.copy()

        # check kwargs
        kw_set = {"weights", "delays", "type", "inh_weight_factor"}

        remaining = set(kwargs) - kw_set

        for kw in remaining:
            _log_message(logger, "WARNING", "Unused keyword argument '" +
                         kw + "'.")

        # update the counters
        self.__class__.__num_graphs += 1
        self.__class__.__max_id += 1

    def __del__(self):
        ''' Graph deletion (update graph count) '''
        self.__class__.__num_graphs -= 1

    def __repr__(self):
        ''' Provide unambiguous informations regarding the object. '''
        d = "directed" if self.is_directed() else "undirected"
        w = "weighted" if self.is_weighted() else "binary"
        t = self.type
        n = self.node_nb()
        e = self.edge_nb()

        return "<{directed}/{weighted} {obj} object of type '{net_type}' " \
               "with {nodes} nodes and {edges} edges at 0x{obj_id}>".format(
                    directed=d, weighted=w, obj=type(self).__name__,
                    net_type=t, nodes=n, edges=e, obj_id=id(self))

    def __str__(self):
        '''
        Return the full string description of the object as would be stored
        inside a file when saving the graph.
        '''
        return _as_string(self)

    @property
    def graph(self):
        '''
        Returns the underlying library object.

        .. warning ::
            Do not add or remove edges directly through this object.

        See also
        --------
        :ref:`graph_attr`
        :ref:`graph-analysis`.
        '''
        return self._graph

    @property
    def structure(self):
        '''
        Object structuring the graph into specific groups.

        .. versionadded: 2.2

        Note
        ----
        Points to :py:obj:`~nngt.Network.population` if the graph is a
        :class:`~nngt.Network`.
        '''
        if self.is_network():
            return self.population

        return self._struct

    @structure.setter
    def structure(self, structure):
        if self.is_network():
            self.population = structure
        else:
            if issubclass(structure.__class__, nngt.Structure):
                if self.node_nb() == structure.size:
                    if structure.is_valid:
                        self._struct = structure
                    else:
                        raise AttributeError(
                            "Structure is not valid (not all nodes are "
                            "associated to a group).")
                else:
                    raise AttributeError("Graph and Structure must have same "
                                         "number of nodes.")
            else:
                raise AttributeError(
                    "Expecting Structure but received '{}'.".format(
                        structure.__class__.__name__))

    @property
    def graph_id(self):
        ''' Unique :class:`int` identifying the instance. '''
        return self.__id

    @property
    def name(self):
        ''' Name of the graph. '''
        return self._name

    @property
    def type(self):
        ''' Type of the graph. '''
        return self._graph_type

    #-------------------------------------------------------------------------#
    # Graph actions

    def copy(self):
        '''
        Returns a deepcopy of the current :class:`~nngt.Graph`
        instance
        '''
        if nngt.get_config("mpi"):
            raise NotImplementedError("`copy` is not MPI-safe yet.")

        gc_instance = Graph(name=self._name + '_copy',
                            weighted=self.is_weighted(), copy_graph=self,
                            directed=self.is_directed())

        if self.is_spatial():
            nngt.SpatialGraph.make_spatial(
                gc_instance, shape=self.shape.copy(),
                positions=deepcopy(self._pos))

        if self.is_network():
            nngt.Network.make_network(gc_instance, self.population.copy())

        return gc_instance

    def to_file(self, filename, fmt="auto", separator=" ", secondary=";",
                attributes=None, notifier="@"):
        '''
        Save graph to file; options detailed below.

        See also
        --------
        :py:func:`nngt.lib.save_to_file` function for options.
        '''
        save_to_file(self, filename, fmt=fmt, separator=separator,
                     secondary=secondary, attributes=attributes,
                     notifier=notifier)

    #~ def inhibitory_subgraph(self):
        #~ ''' Create a :class:`~nngt.Graph` instance which graph
        #~ contains only the inhibitory edges of the current instance's
        #~ :class:`graph_tool.Graph` '''
        #~ eprop_b_type = self.new_edge_property(
                       #~ "bool",-self.edge_properties[TYPE].a+1)
        #~ self.set_edge_filter(eprop_b_type)
        #~ inhib_graph = Graph( name=self._name + '_inhib',
                             #~ weighted=self._weighted,
                             #~ from_graph=core.GraphObject(self.prune=True) )
        #~ self.clear_filters()
        #~ return inhib_graph

    #~ def excitatory_subgraph(self):
        #~ '''
        #~ Create a :class:`~nngt.Graph` instance which graph contains only the
        #~ excitatory edges of the current instance's :class:`core.GraphObject`.
        #~ .. warning ::
            #~ Only works for graph_tool
        #~ .. todo ::
            #~ Make this method library independant!
        #~ '''
        #~ eprop_b_type = self.new_edge_property(
                       #~ "bool",self.edge_properties[TYPE].a+1)
        #~ self.set_edge_filter(eprop_b_type)
        #~ exc_graph = Graph( name=self._name + '_exc',
                             #~ weighted=self._weighted,
                             #~ graph=core.GraphObject(self.prune=True) )
        #~ self.clear_filters()
        #~ return exc_graph

    def to_undirected(self, combine_numeric_eattr="sum"):
        '''
        Convert the graph to its undirected variant.

        .. note::
            All non-numeric edge attributes will be discarded from the returned
            undirected graph.

        Parameters
        ----------
        combine_numeric_eattr : str, optional (default: "sum")
            How to combine numeric attributes from reciprocal edges.
            Can be either:

            - "sum" (attributes are summed)
            - "min" (smallest value is kept)
            - "max" (largest value is kept)
            - "mean" (the average of both attributes is taken)

            In addition, `combine_numeric_eattr` can be a dictionary with one
            entry for each edge attribute.
        '''
        shape = self.shape if self.is_spatial() else None
        pos   = self.get_positions() if self.is_spatial() else None

        g = self.__class__(self.node_nb(), structure=self.structure,
                           weighted=self.is_weighted(), directed=False)

        if shape is not None or pos is not None:
            g.make_spatial(g, shape=shape, positions=pos)

        # replicate node attributes
        for nattr in self.node_attributes:
            g.new_node_attribute(nattr, self.get_attribute_type(nattr, "node"),
                                 self.node_attributes[nattr])

        # prepare edges
        eattrs = set(self.edge_attributes)

        # prepare combine method
        if isinstance(combine_numeric_eattr, str):
            val = str(combine_numeric_eattr)
            combine_numeric_eattr = defaultdict(lambda: val)
        elif isinstance(combine_numeric_eattr, dict):
            combine_numeric_eattr = defaultdict(
                lambda: "sum", **combine_numeric_eattr)

        # find integer eattr
        numeric_eattr = "weight" if "weight" in eattrs else None
        numeric_types = ("int", "double")

        if numeric_eattr is None:
            for eattr in eattrs:
                if self.get_attribute_type(eattr, "edge") in numeric_types:
                    numeric_eattr = eattr
                    break

        if numeric_eattr is not None:
            eattrs.discard(numeric_eattr)

            combine = combine_numeric_eattr[numeric_eattr]

            _, umat = _get_matrices(
                self, directed=False, weights=numeric_eattr, weighted=True,
                combine_weights=combine, remove_self_loops=False)

            umat = ssp.tril(umat, format="csr")

            # create the initial edge attribute
            g.new_edge_attribute(
                numeric_eattr, self.get_attribute_type(numeric_eattr, "edge"))

            indptr = umat.indptr

            diff = np.diff(indptr)
            keep = np.where(diff)[0]
            sources = np.repeat(keep, diff[keep])

            # make and add the edges and the first eattr
            edges = np.array((sources, umat.indices)).T

            g.new_edges(edges, attributes={"weight": umat.data},
                        check_self_loops=False)

            # add all other edge attributes
            for eattr in eattrs:
                etype = self.get_attribute_type(eattr, "edge")

                combine = combine_numeric_eattr[eattr]

                if etype in numeric_types:
                    if np.all(self.edge_attributes[eattr] > 0):
                        _, umat = _get_matrices(
                            self, directed=False, weights=eattr, weighted=True,
                            combine_weights=combine, remove_self_loops=False)

                        umat = ssp.tril(umat, format="csr")

                        g.new_edge_attribute(
                            eattr, self.get_attribute_type(eattr, "edge"),
                            values=umat.data)
                    else:
                        aa = list(self.edge_attributes[eattr])

                        adict = {
                            tuple(e): val
                            for e, val in zip(self.edges_array, aa)
                        }

                        f = None

                        if combine == "max":
                            f = np.max
                        elif combine == "min":
                            f = np.min
                        elif combine == "mean":
                            f = np.mean
                        elif combine == "sum":
                            f = np.sum
                        else:
                            raise ValueError(
                                "Invalid combination mode '{}'.".format(
                                    combine))

                        values = [
                            f([adict[e] for e in {tuple(e0), tuple(e0[::-1])}
                               if e in adict]) for e0 in g.edges_array
                        ]

                        g.new_edge_attribute(
                            eattr, self.get_attribute_type(eattr, "edge"),
                            values=values)
        else:
            # hide existing edge warning
            from nngt.lib.connect_tools import logger as lg

            old_loglevel = lg.level
            lg.setLevel(logging.ERROR)

            g.new_edges(self.edges_array, ignore_invalid=True)

            # restore previous logging level
            lg.setLevel(old_loglevel)

        return g


    def get_structure_graph(self):
        '''
        Return a coarse-grained version of the graph containing one node
        per :class:`nngt.Group`.
        Connections between groups are associated to the sum of all connection
        weights.
        If no structure is present, returns an empty Graph.
        '''
        struct = self.structure

        if struct is None:
            return Graph()

        names = list(struct.keys())
        nodes = len(struct)

        g = nngt.Graph(nodes,
                       name="Structure-graph of '{}'".format(self.name))

        eattr = {"weight": []}

        if self.is_network():
            eattr["delay"] = []

        new_edges = []

        for i, n1 in enumerate(names):
            g1 = struct[n1]

            for j, n2 in enumerate(names):
                g2 = struct[n2]

                edges = self.get_edges(source_node=g1.ids, target_node=g2.ids)

                if len(edges):
                    weights = self.get_weights(edges=edges)
                    w = np.sum(weights)
                    eattr["weight"].append(w)

                    if self.is_network():
                        delays  = self.get_delays(edges=edges)
                        d = np.average(delays)
                        eattr["delay"].append(d)

                    new_edges.append((i, j))

        # add edges and attributes
        if self.is_network():
            g.new_edge_attribute("delay", "double")

        g.new_edges(new_edges, attributes=eattr, check_self_loops=False)

        # set node attributes
        g.new_node_attribute("name", "string", values=names)

        return g

    #-------------------------------------------------------------------------#
    # Getters

    def adjacency_matrix(self, types=False, weights=False, mformat="csr"):
        '''
        Return the graph adjacency matrix.

        .. versionchanged: 2.0
            Added matrix format option (`mformat`).

        Note
        ----
        Source nodes are represented by the rows, targets by the
        corresponding columns.

        Parameters
        ----------
        types : bool, optional (default: False)
            Wether the edge types should be taken into account (negative values
            for inhibitory connections).
        weights : bool or string, optional (default: False)
            Whether the adjacecy matrix should be weighted. If True, all
            connections are multiply bythe associated synaptic strength; if
            weight is a string, the connections are scaled bythe corresponding
            edge attribute.
        mformat : str, optional (default: "csr")
            Type of :mod:`scipy.sparse` matrix that will be returned, by
            default :class:`scipy.sparse.csr_matrix`.

        Returns
        -------
        mat : :mod:`scipy.sparse` matrix
            The adjacency matrix of the graph.
        '''
        weights = "weight" if weights is True else weights

        mat = None

        if types:
            if self.is_network():
                # use inhibitory nodes
                mat = nngt.analyze_graph["adjacency"](self, weights)
                inh = self.population.inhibitory

                if np.any(inh):
                    mat[inh, :] *= -1

            elif 'type' in self.node_attributes:
                mat = nngt.analyze_graph["adjacency"](self, weights)
                tarray = np.where(self.node_attributes['type'] < 0)[0]
                if np.any(tarray):
                    mat[tarray] *= -1
            elif types and 'type' in self.edge_attributes:
                data = None

                if nonstring_container(weights):
                    data = weights
                elif weights in {None, False}:
                    data = np.ones(self.edge_nb())
                else:
                    data = self.get_edge_attributes(name=weights)

                data *= self.get_edge_attributes(name="type")

                edges     = self.edges_array
                num_nodes = self.node_nb()
                mat       = ssp.coo_matrix(
                    (data, (edges[:, 0], edges[:, 1])),
                    shape=(num_nodes, num_nodes)).tocsr()

                if not self.is_directed():
                    mat += mat.T

            return mat.asformat(mformat)

        # untyped
        mat = nngt.analyze_graph["adjacency"](self, weights, mformat=mformat)

        return mat

    @property
    def node_attributes(self):
        '''
        Access node attributes.

        See also
        --------
        :attr:`~nngt.Graph.edge_attributes`,
        :attr:`~nngt.Graph.get_node_attributes`,
        :attr:`~nngt.Graph.new_node_attribute`,
        :attr:`~nngt.Graph.set_node_attribute`.
        '''
        return self._nattr

    @property
    def edge_attributes(self):
        '''
        Access edge attributes.

        See also
        --------
        :attr:`~nngt.Graph.node_attributes`,
        :attr:`~nngt.Graph.get_edge_attributes`,
        :attr:`~nngt.Graph.new_edge_attribute`,
        :attr:`~nngt.Graph.set_edge_attribute`.
        '''
        return self._eattr

    def get_nodes(self, attribute=None, value=None):
        '''
        Return the nodes in the network fulfilling a given condition.

        Parameters
        ----------
        attribute : str, optional (default: all nodes)
            Whether the `attribute` of the returned nodes should have a specific
            value.
        value : object, optional (default : None)
            If an `attribute` name is passed, then only nodes with `attribute`
            being equal to `value` will be returned.

        See also
        --------
        :func:`~nngt.Graph.get_edges`, :attr:`~nngt.Graph.node_attributes`
        '''
        if attribute is None:
            return [i for i in range(self.node_nb())]

        vtype = self._nattr.value_type(attribute)

        if value is None and vtype != "object":
            raise ValueError("`value` cannot be None for attribute '" +
                             attribute + "'.")

        return np.where(
            self.get_node_attributes(name=attribute) == value)[0]

    def get_edges(self, attribute=None, value=None, source_node=None,
                  target_node=None):
        '''
        Return the edges in the network fulfilling a given condition.

        For undirected graphs, edges are always returned in the order
        :math:`(u, v)` where :math:`u <= v`.

        .. warning ::
            Contrary to :func:`~nngt.Graph.edges_array` that returns edges
            ordered by creation time (i.e. corresponding to the order of the
            edge attribute array), this function does not enforce any specific
            edge order.
            This also  means that, if order does not matter, it may be faster
            to call ``get_edges`` that to call ``edges_array``.

        Parameters
        ----------
        attribute : str, optional (default: all nodes)
            Whether the `attribute` of the returned edges should have a
            specific value.
        value : object, optional (default : None)
            If an `attribute` name is passed, then only edges with `attribute`
            being equal to `value` will be returned.
        source_node : int or list of ints, optional (default: all nodes)
            Retrict the edges to those stemming from `source_node`.
        target_node : int or list of ints, optional (default: all nodes)
            Retrict the edges to those arriving at `target_node`.

        Returns
        -------
        A list of edges (2-tuples).

        See also
        --------
        :func:`~nngt.Graph.get_nodes`, :attr:`~nngt.Graph.edge_attributes`,
        :func:`~nngt.Graph.edges_array`
        '''
        edges = None

        if is_integer(source_node) and is_integer(target_node):
            # check that the edge exists, throw error otherwise
            self.edge_id((source_node, target_node))
            edges = [(source_node, target_node)]
        else:
            # backend-specific implementation for source or target
            edges = self._get_edges(source_node=source_node,
                                    target_node=target_node)

        # check attributes
        if attribute is None:
            return edges

        vtype = self._eattr.value_type(attribute)

        if value is None and vtype != "object":
            raise ValueError("`value` cannot be None for attribute '" +
                             attribute + "'.")

        desired = (self.get_edge_attributes(edges, attribute) == value)

        return [tuple(e) for e in self.edges_array[desired]]

    def get_edge_attributes(self, edges=None, name=None):
        '''
        Attributes of the graph's edges.

        .. versionchanged:: 1.0
            Returns the full dict of edges attributes if called without
            arguments.

        .. versionadded:: 0.8

        Parameters
        ----------
        edge : tuple or list of tuples, optional (default: ``None``)
            Edge whose attribute should be displayed.
        name : str, optional (default: ``None``)
            Name of the desired attribute.

        Returns
        -------
        Dict containing all graph's attributes (synaptic weights, delays...)
        by default. If `edge` is specified, returns only the values for these
        edges. If `name` is specified, returns value of the attribute for each
        edge.

        Note
        ----
        The attributes values are ordered as the edges in
        :func:`~nngt.Graph.edges_array` if `edges` is None.

        See also
        --------
        :func:`~nngt.Graph.get_node_attributes`,
        :func:`~nngt.Graph.new_edge_attribute`,
        :func:`~nngt.Graph.set_edge_attribute`,
        :func:`~nngt.Graph.new_node_attribute`,
        :func:`~nngt.Graph.set_node_attribute`
        '''
        if name is not None and edges is not None:
            if isinstance(edges, slice):
                return self._eattr[name][edges]
            elif len(edges):
                return self._eattr[edges][name]
            return np.array([])
        elif name is None and edges is None:
            return {k: self._eattr[k]
                    for k in self._eattr.keys()}
        elif name is None:
            return self._eattr[edges]
        else:
            return self._eattr[name]

    def get_node_attributes(self, nodes=None, name=None):
        '''
        Attributes of the graph's edges.

        .. versionchanged:: 1.0.1
            Corrected default behavior and made it the same as
            :func:`~nngt.Graph.get_edge_attributes`.

        .. versionadded:: 0.9

        Parameters
        ----------
        nodes : list of ints, optional (default: ``None``)
            Nodes whose attribute should be displayed.
        name : str, optional (default: ``None``)
            Name of the desired attribute.

        Returns
        -------
        Dict containing all nodes attributes by default. If `nodes` is
        specified, returns a ``dict`` containing only the attributes of these
        nodes. If `name` is specified, returns a list containing the values of
        the specific attribute for the required nodes (or all nodes if
        unspecified).

        See also
        --------
        :func:`~nngt.Graph.get_edge_attributes`,
        :func:`~nngt.Graph.new_node_attribute`,
        :func:`~nngt.Graph.set_node_attribute`,
        :func:`~nngt.Graph.new_edge_attributes`,
        :func:`~nngt.Graph.set_edge_attribute`
        '''
        res = None

        if name is None:
            res = {k: self._nattr[k] for k in self._nattr.keys()}
        else:
            res = self._nattr[name]

        if nodes is None:
            return res

        if isinstance(nodes, (slice, int)) or nonstring_container(nodes):
            if isinstance(res, dict):
                return {k: v[nodes] for k, v in res.items()}
            return res[nodes]
        else:
            raise ValueError("Invalid `nodes`: "
                             "{}, use slice, int, or list".format(nodes))

    def get_attribute_type(self, attribute_name, attribute_class=None):
        '''
        Return the type of an attribute (e.g. string, double, int).

        Parameters
        ----------
        attribute_name : str
            Name of the attribute.
        attribute_class : str, optional (default: both)
            Whether `attribute_name` is a "node" or an "edge" attribute.

        Returns
        -------
        type : str
            Type of the attribute.
        '''
        if attribute_class is None:
            is_eattr = attribute_name in self._eattr
            is_nattr = attribute_name in self._nattr
            if is_eattr and is_nattr:
                raise RuntimeError("Both edge and node attributes with name '"
                                   + attribute_name + "' exist, please "
                                   "specify `attribute_class`")
            elif is_eattr:
                return self._eattr.value_type(attribute_name)
            elif is_nattr:
                return self._nattr.value_type(attribute_name)
            else:
                raise KeyError("No '{}' attribute.".format(attribute_name))
        else:
            if attribute_class == "edge":
                return self._eattr.value_type(attribute_name)
            elif attribute_class == "node":
                return self._nattr.value_type(attribute_name)
            else:
                raise InvalidArgument(
                    "Unknown attribute class '{}'.".format(attribute_class))

    def get_density(self):
        '''
        Density of the graph: :math:`\\frac{E}{N^2}`, where `E` is the number
        of edges and `N` the number of nodes.
        '''
        return self.edge_nb() / self.node_nb()**2

    def is_weighted(self):
        ''' Whether the edges have weights '''
        return "weight" in self.edge_attributes

    def is_directed(self):
        ''' Whether the graph is directed or not '''
        return self._graph.is_directed()

    def is_connected(self, mode="strong"):
        '''
        Return whether the graph is connected.

        Parameters
        ----------
        mode : str, optional (default: "strong")
            Whether to test connectedness with directed ("strong") or
            undirected ("weak") connections.

        References
        ----------
        .. [ig-connected] :igdoc:`is_connected`
        '''
        return super().is_connected()

    def get_degrees(self, mode="total", nodes=None, weights=None,
                    edge_type="all"):
        '''
        Degree sequence of all the nodes.

        .. versionchanged:: 2.0
            Changed `deg_type` to `mode`, `node_list` to `nodes`, `use_weights`
            to `weights`, and `edge_type` to `edge_type`.

        Parameters
        ----------
        mode : string, optional (default: "total")
            Degree type (among 'in', 'out' or 'total').
        nodes : list, optional (default: None)
            List of the nodes which degree should be returned
        weights : bool or str, optional (default: binary edges)
            Whether edge weights should be considered; if ``None`` or ``False``
            then use binary edges; if ``True``, uses the 'weight' edge
            attribute, otherwise uses any valid edge attribute required.
        edge_type : int or str, optional (default: all)
            Restrict to a given synaptic type ("excitatory", 1, or
            "inhibitory", -1), using either the "type" edge attribute for
            non-:class:`~nngt.Network` or the
            :py:attr:`~nngt.NeuralPop.inhibitory` nodes.

        Returns
        -------
        degrees : :class:`numpy.array`
        
        .. warning ::
            When using MPI with "nngt" (distributed) backend, returns only the
            degrees associated to local edges. "Complete" degrees are obtained
            by taking the sum of the results on all MPI processes.
        '''
        mode = _set_degree_type(mode)

        if edge_type == "all":
            return super().get_degrees(
                mode=mode, nodes=nodes, weights=weights)
        elif edge_type in {"excitatory", 1}:
            edge_type = 1
        elif edge_type in {"inhibitory", -1}:
            edge_type = -1
        else:
            raise InvalidArgument(
                "Invalid edge type '{}'".format(edge_type))

        degrees = np.zeros(self.node_nb())

        if isinstance(self, nngt.Network):
            neurons = []
            for g in self.population.values():
                if g.neuron_type == edge_type:
                    neurons.extend(g.ids)

            if mode in {"in", "all"} or not self.is_directed():
                degrees += self.adjacency_matrix(
                    weights=weights,
                    types=False)[neurons, :].sum(axis=0).A1

            if mode in {"out", "all"} and self.is_directed():
                degrees += self.adjacency_matrix(
                    weights=weights,
                    types=False)[neurons, :].sum(axis=1).A1
        else:
            edges = np.where(
                self.get_edge_attributes(name="type") == edge_type)[0]

            w = None

            if weights is None:
                w = np.ones(len(edges))
            elif weights in self.edge_attributes:
                w = self.edge_attributes[weights]
            elif nonstring_container(weights):
                w = np.array(weights)
            else:
                raise InvalidArgument(
                    "Invalid `weights` '{}'".format(weights))

            # count in-degrees
            if mode in {"in", "all"} or not self.is_directed():
                np.add.at(degrees, edges[1], weights)

            if mode in {"out", "all"} and self.is_directed():
                np.add.at(degrees, edges[0], weights)

        if nodes is None:
            return degrees

        return degrees[nodes]

    def get_betweenness(self, btype="both", weights=None):
        '''
        Returns the normalized betweenness centrality of the nodes and edges.

        Parameters
        ----------
        g : :class:`~nngt.Graph`
            Graph to analyze.
        btype : str, optional (default 'both')
            The centrality that should be returned (either 'node', 'edge', or
            'both'). By default, both betweenness centralities are computed.
        weights : bool or str, optional (default: binary edges)
            Whether edge weights should be considered; if ``None`` or
            ``False`` then use binary edges; if ``True``, uses the 'weight'
            edge attribute, otherwise uses any valid edge attribute required.

        Returns
        -------
        nb : :class:`numpy.ndarray`
            The nodes' betweenness if `btype` is 'node' or 'both'
        eb : :class:`numpy.ndarray`
            The edges' betweenness if `btype` is 'edge' or 'both'

        See also
        --------
        :func:`~nngt.analysis.betweenness`
        '''
        from nngt.analysis import betweenness
        return betweenness(self, btype=btype, weights=weights)

    def get_edge_types(self, edges=None):
        '''
        Return the type of all or a subset of the edges.

        Parameters
        ----------
        edges : (E, 2) array, optional (default: all edges)
            Edges for which the type should be returned.

        Returns
        -------
        the list of types (1 for excitatory, -1 for inhibitory)
        '''
        if TYPE in self.edge_attributes:
            return self.get_edge_attributes(name=TYPE, edges=edges)
        else:
            size = self.edge_nb() if edges is None else len(edges)
            return np.ones(size)

    def get_weights(self, edges=None):
        '''
        Returns the weights of all or a subset of the edges.

        .. versionchanged:: 1.0.1
            Added the possibility to ask for a subset of edges.

        Parameters
        ----------
        edges : (E, 2) array, optional (default: all edges)
            Edges for which the type should be returned.

        Returns
        -------
        the list of weights
        '''
        if self.is_weighted():
            if edges is None:
                return self._eattr["weight"]
            else:
                if len(edges) == 0:
                    return np.array([])

                return np.asarray(self._eattr[edges]["weight"])
        else:
            size = self.edge_nb() if edges is None else len(edges)
            return np.ones(size)

    def get_delays(self, edges=None):
        '''
        Returns the delays of all or a subset of the edges.

        .. versionchanged:: 1.0.1
            Added the possibility to ask for a subset of edges.

        Parameters
        ----------
        edges : (E, 2) array, optional (default: all edges)
            Edges for which the type should be returned.

        Returns
        -------
        the list of delays
        '''
        if edges is None:
            return self._eattr["delay"]
        else:
            return self._eattr[edges]["delay"]

    def neighbours(self, node, mode="all"):
        '''
        Return the neighbours of `node`.

        Parameters
        ----------
        node : int
            Index of the node of interest.
        mode : string, optional (default: "all")
            Type of neighbours that will be returned: "all" returns all the
            neighbours regardless of directionality, "in" returns the
            in-neighbours (also called predecessors) and "out" retruns the
            out-neighbours (or successors).

        Returns
        -------
        neighbours : set
            The neighbours of `node`.
        '''
        return super().neighbours(node, mode=mode)

    def is_spatial(self):
        '''
        Whether the graph is embedded in space (i.e. is a subclass of
        :class:`~nngt.SpatialGraph`).
        '''
        return issubclass(self.__class__, nngt.SpatialGraph)

    def is_network(self):
        '''
        Whether the graph is a subclass of :class:`~nngt.Network` (i.e. if it
        has a :class:`~nngt.NeuralPop` attribute).
        '''
        return issubclass(self.__class__, nngt.Network)

    #-------------------------------------------------------------------------#
    # Setters

    def set_name(self, name=""):
        ''' set graph name '''
        if name != "":
            self._name = name
        else:
            self._name = "Graph_" + str(self.__id)

    def new_edge_attribute(self, name, value_type, values=None, val=None):
        '''
        Create a new attribute for the edges.

        Parameters
        ----------
        name : str
            The name of the new attribute.
        value_type : str
            Type of the attribute, among 'int', 'double', 'string', or 'object'
        values : array, optional (default: None)
            Values with which the edge attribute should be initialized.
            (must have one entry per node in the graph)
        val : int, float or str , optional (default: None)
            Identical value for all edges.
        '''
        assert name != "eid", "`eid` is a reserved internal edge-attribute."

        self._eattr.new_attribute(
            name, value_type, values=values, val=val)

    def new_node_attribute(self, name, value_type, values=None, val=None):
        '''
        Create a new attribute for the nodes.

        Parameters
        ----------
        name : str
            The name of the new attribute.
        value_type : str
            Type of the attribute, among 'int', 'double', 'string', or 'object'
        values : array, optional (default: None)
            Values with which the node attribute should be initialized.
            (must have one entry per node in the graph)
        val : int, float or str , optional (default: None)
            Identical value for all nodes.

        See also
        --------
        :func:`~nngt.Graph.new_edge_attribute`,
        :func:`~nngt.Graph.set_node_attribute`,
        :func:`~nngt.Graph.get_node_attributes`,
        :func:`~nngt.Graph.set_edge_attribute`,
        :func:`~nngt.Graph.get_edge_attributes`
        '''
        self._nattr.new_attribute(
            name, value_type, values=values, val=val)

    def set_edge_attribute(self, attribute, values=None, val=None,
                           value_type=None, edges=None):
        '''
        Set attributes to the connections between neurons.

        .. warning ::
            The special "type" attribute cannot be modified when using graphs
            that inherit from the :class:`~nngt.Network` class. This is because
            for biological networks, neurons make only one kind of synapse,
            which is determined by the :class:`nngt.NeuralGroup` they
            belong to.

        Parameters
        ----------
        attribute : str
            The name of the attribute.
        value_type : str
            Type of the attribute, among 'int', 'double', 'string'
        values : array, optional (default: None)
            Values with which the edge attribute should be initialized.
            (must have one entry per node in the graph)
        val : int, float or str , optional (default: None)
            Identical value for all edges.
        value_type : str, optional (default: None)
            Type of the attribute, among 'int', 'double', 'string'. Only used
            if the attribute does not exist and must be created.
        edges : list of edges or array of shape (E, 2), optional (default: all)
            Edges whose attributes should be set. Others will remain unchanged.

        See also
        --------
        :func:`~nngt.Graph.set_node_attribute`,
        :func:`~nngt.Graph.get_edge_attributes`,
        :func:`~nngt.Graph.new_edge_attribute`,
        :func:`~nngt.Graph.new_node_attribute`,
        :func:`~nngt.Graph.get_node_attributes`
        '''
        if attribute not in self.edge_attributes:
            assert value_type is not None, "`value_type` is necessary for " +\
                                           "new attributes."
            self.new_edge_attribute(name=attribute, value_type=value_type,
                                    values=values, val=val)
        else:
            num_edges = self.edge_nb() if edges is None else len(edges)
            if values is None:
                if val is not None:
                    values = [deepcopy(val) for _ in range(num_edges)]
                else:
                    raise InvalidArgument("At least one of the `values` and "
                        "`val` arguments should not be ``None``.")
            self._eattr.set_attribute(attribute, values, edges=edges)

    def set_node_attribute(self, attribute, values=None, val=None,
                           value_type=None, nodes=None):
        '''
        Set attributes to the connections between neurons.

        Parameters
        ----------
        attribute : str
            The name of the attribute.
        value_type : str
            Type of the attribute, among 'int', 'double', 'string'
        values : array, optional (default: None)
            Values with which the edge attribute should be initialized.
            (must have one entry per node in the graph)
        val : int, float or str , optional (default: None)
            Identical value for all edges.
        value_type : str, optional (default: None)
            Type of the attribute, among 'int', 'double', 'string'. Only used
            if the attribute does not exist and must be created.
        nodes : list of nodes, optional (default: all)
            Nodes whose attributes should be set. Others will remain unchanged.

        See also
        --------
        :func:`~nngt.Graph.set_edge_attribute`,
        :func:`~nngt.Graph.new_node_attribute`,
        :func:`~nngt.Graph.get_node_attributes`,
        :func:`~nngt.Graph.new_edge_attribute`,
        :func:`~nngt.Graph.get_edge_attributes`,
        '''
        if attribute not in self.node_attributes:
            assert value_type is not None, "`value_type` is necessary for " +\
                                           "new attributes."
            self.new_node_attribute(name=attribute, value_type=value_type,
                                    values=values, val=val)
        else:
            num_nodes = self.node_nb() if nodes is None else len(nodes)
            if values is None:
                if val is not None:
                    values = [deepcopy(val) for _ in  range(num_nodes)]
                else:
                    raise InvalidArgument("At least one of the `values` and "
                        "`val` arguments should not be ``None``.")
            self._nattr.set_attribute(attribute, values, nodes=nodes)

    def set_weights(self, weight=None, elist=None, distribution=None,
                    parameters=None, noise_scale=None):
        '''
        Set the synaptic weights.

        Parameters
        ----------
        weight : float or class:`numpy.array`, optional (default: None)
            Value or list of the weights (for user defined weights).
        elist : class:`numpy.array`, optional (default: None)
            List of the edges (for user defined weights).
        distribution : class:`string`, optional (default: None)
            Type of distribution (choose among "constant", "uniform",
            "gaussian", "lognormal", "lin_corr", "log_corr").
        parameters : dict, optional (default: {})
            Dictionary containing the properties of the weight distribution.
            Properties are as follow for the distributions

            - 'constant': 'value'
            - 'uniform': 'lower', 'upper'
            - 'gaussian': 'avg', 'std'
            - 'lognormal': 'position', 'scale'

        noise_scale : class:`int`, optional (default: None)
            Scale of the multiplicative Gaussian noise that should be applied
            on the weights.

        Note
        ----
        If `distribution` and `parameters` are provided and the weights are set
        for the whole graph (`elist` is None), then the distribution properties
        will be kept as the new default for subsequent edges. That is, if new
        edges are created without specifying their weights, then these new
        weights will automatically be drawn from this previous distribution.
        '''
        if isinstance(weight, float):
            size = self.edge_nb() if elist is None else len(elist)
            self._w = {"distribution": "constant", "value": weight}
            weight = np.repeat(weight, size)
        elif not nonstring_container(weight) and weight is not None:
            raise AttributeError("Invalid `weight` value: must be either "
                                 "float, array-like or None.")
        elif weight is not None:
            self._w = {"distribution": "custom"}
        elif None not in (distribution, parameters) and elist is None:
            self._w = {"distribution": distribution}
            self._w.update(parameters)

        if distribution is None:
            distribution = self._w.get("distribution", None)

        if parameters is None:
            parameters = self._w

        Connections.weights(
            self, elist=elist, wlist=weight, distribution=distribution,
            parameters=parameters, noise_scale=noise_scale)

    def set_types(self, edge_type, nodes=None, fraction=None):
        '''
        Set the synaptic/connection types.

        .. versionchanged :: 2.0
            Changed `syn_type` to `edge_type`.

        .. warning ::
            The special "type" attribute cannot be modified when using graphs
            that inherit from the :class:`~nngt.Network` class. This is because
            for biological networks, neurons make only one kind of synapse,
            which is determined by the :class:`nngt.NeuralGroup` they
            belong to.

        Parameters
        ----------
        edge_type : int, string, or array of ints
            Type of the connection among 'excitatory' (also `1`) or
            'inhibitory' (also `-1`).
        nodes : int, float or list, optional (default: `None`)
            If `nodes` is an int, number of nodes of the required type that
            will be created in the graph (all connections from inhibitory nodes
            are inhibitory); if it is a float, ratio of `edge_type` nodes in the
            graph; if it is a list, ids of the `edge_type` nodes.
        fraction : float, optional (default: `None`)
            Fraction of the selected edges that will be set as `edge_type` (if
            `nodes` is not `None`, it is the fraction of the specified nodes'
            edges, otherwise it is the fraction of all edges in the graph).

        Returns
        -------
        t_list : :class:`numpy.ndarray`
            List of the types in an order that matches the `edges` attribute of
            the graph.
        '''
        inhib_nodes = None

        if nonstring_container(edge_type):
            return Connections.types(self, values=edge_type)
        elif edge_type in ('excitatory', 1):
            if is_integer(nodes):
                inhib_nodes = self.node_nb() - nodes
            elif nonstring_container(nodes):
                inhib_nodes = list(range(self.node_nb()))
                nodes.sort()
                for node in nodes[::-1]:
                    del inhib_nodes[node]
            elif nodes is not None:
                raise ValueError("`nodes` should be integer or array of ids.")
        elif edge_type in ('inhibitory', -1):
            if is_integer(nodes) or nonstring_container(nodes):
                inhib_nodes = nodes
            elif nodes is not None:
                raise ValueError("`nodes` should be integer or array of ids.")

        return Connections.types(self, inhib_nodes, fraction)

    def set_delays(self, delay=None, elist=None, distribution=None,
                   parameters=None, noise_scale=None):
        '''
        Set the delay for spike propagation between neurons.

        Parameters
        ----------
        delay : float or class:`numpy.array`, optional (default: None)
            Value or list of delays (for user defined delays).
        elist : class:`numpy.array`, optional (default: None)
            List of the edges (for user defined delays).
        distribution : class:`string`, optional (default: None)
            Type of distribution (choose among "constant", "uniform",
            "gaussian", "lognormal", "lin_corr", "log_corr").
        parameters : dict, optional (default: {})
            Dictionary containing the properties of the delay distribution.
        noise_scale : class:`int`, optional (default: None)
            Scale of the multiplicative Gaussian noise that should be applied
            on the delays.
        '''
        # check special cases and set self._d
        if isinstance(delay, float):
            size = self.edge_nb() if elist is None else len(elist)
            self._d = {"distribution": "constant", "value": delay}
            delay = np.repeat(delay, size)
        elif not nonstring_container(delay) and delay is not None:
            raise AttributeError("Invalid `delay` value: must be either "
                                 "float, array-like or None")
        elif delay is not None:
            self._d = {"distribution": "custom"}
        elif None not in (distribution, parameters):
            self._d = {"distribution": distribution}
            self._d.update(parameters)

        if distribution is None:
            if hasattr(self, "_d"):
                distribution = self._d["distribution"]
            else:
                raise AttributeError(
                    "Invalid `distribution` value: cannot be None if "
                    "default delays were not set at graph creation.")

        if parameters is None:
            if hasattr(self, "_d"):
                parameters = self._d
            else:
                raise AttributeError(
                    "Invalid `parameters` value: cannot be None if default"
                    " delays were not set at graph creation.")

        return Connections.delays(
            self, elist=elist, dlist=delay, distribution=distribution,
            parameters=parameters, noise_scale=noise_scale)



#-*- coding:utf-8 -*-
#
# lib/rng_tools.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Generating the weights of the graph object's connections """

import numpy as np
import scipy.sparse as ssp

import nngt

from .errors import InvalidArgument
from .test_functions import nonstring_container
from .test_functions import mpi_random


# ----------- #
# Random seed #
# ----------- #

@mpi_random
def seed(msd=None, seeds=None):
    '''
    Seed the random generator used by NNGT
    (i.e. the numpy `RandomState`: for details, see
    :class:`numpy.random.RandomState`).

    Parameters
    ----------
    msd : int, optional
        Master seed for numpy `RandomState`.
        Must be convertible to 32-bit unsigned integers.
    seeds : list of ints, optional
        Seeds for `RandomState` (when using MPI).
        Must be convertible to 32-bit unsigned integers, one entry per MPI
        process.
    '''
    # when using MPI numpy seeeds are sync-ed via the mpi_random decorator
    msd = np.random.randint(0, 2**31 - 1) if msd is None else msd

    # seed both random state and new generator
    np.random.seed(msd)
    nngt._rng = np.random.default_rng(msd)

    nngt._config['msd'] = msd

    nngt._seeded = True

    nngt._seeded_local = False

    # check subseeds
    if seeds is not None:
        with_mt = nngt.get_config('multithreading')
        with_mpi = nngt.get_config('mpi')
        err = 'Expected {} seeds.'

        if with_mpi:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            size = comm.Get_size()
            assert size == len(seeds), err.format(size)
            nngt._config['seeds'] = seeds
        elif with_mt:
            num_omp = nngt.get_config('omp')
            assert num_omp == len(seeds), err.format(num_omp)
            nngt._config['seeds'] = seeds

        nngt._seeded_local = True
        nngt._used_local   = False


# ----------------------------- #
# Return the right distribution #
# ----------------------------- #

def _generate_random(number, instructions):
    name = "not defined"

    if isinstance(instructions, dict):
        name = instructions["distribution"]

        instructions = {
            k: v for k, v in instructions.items() if k != "distribution"
        }

        if name in di_dfunc:
            return di_dfunc[name](None, None, number, **instructions)

        raise NotImplementedError(
            "Unknown distribution: '{}'. Supported distributions " \
            "are {}".format(name, ", ".join(di_dfunc.keys())))
    elif nonstring_container(instructions):
        name = instructions[0]

        if name in di_dfunc:
            return di_dfunc[name](None, None, number, *instructions[1:])

        raise NotImplementedError(
            "Unknown distribution: '{}'. Supported distributions " \
            "are {}".format(name, ", ".join(di_dfunc.keys())))

    raise NotImplementedError(
        "Unknown instructions: '{}'".format(instructions))


def _eprop_distribution(graph, distrib_type, matrix=False, elist=None,
                        last_edges=False, **kw):
    ra_values = di_dfunc[distrib_type](graph, elist=elist,
                                       last_edges=last_edges, **kw)
    num_edges = graph.edge_nb()

    if matrix:
        return _make_matrix(graph, num_edges, ra_values, elist)
    else:
        return ra_values


# --------------------- #
# Generating the matrix #
# --------------------- #

def _make_matrix(graph, ecount, values, elist=None):
    mat_distrib = None
    n = graph.node_nb()
    if elist is not None and graph.edge_nb():
        mat_distrib = ssp.coo_matrix((values,(elist[:,0],elist[:,1])),(n,n))
    else:
        mat_distrib = graph.adjacency_matrix()
        mat_distrib.data = values
    mat_distrib = mat_distrib.tolil()
    mat_distrib.setdiag(np.zeros(n))
    return mat_distrib


# ----------------------- #
# Distribution generators #
# ----------------------- #

def delta_distrib(graph=None, elist=None, num=None, value=1., **kwargs):
    '''
    Delta distribution for edge attributes.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Graph for which an edge attribute will be generated.
    elist : list of edges, optional (default: all edges)
        Generate values for only a subset of edges.
    value : float, optional (default: 1.)
        Value of the delta distribution.

    Returns : :class:`numpy.ndarray`
        Attribute value for each edge in `graph`.
    '''
    num = _compute_num_prop(elist, graph, num)
    return np.repeat(value, num)


def uniform_distrib(graph, elist=None, num=None, lower=None, upper=None,
                    **kwargs):
    '''
    Uniform distribution for edge attributes.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Graph for which an edge attribute will be generated.
    elist : list of edges, optional (default: all edges)
        Generate values for only a subset of edges.
    lower : float, optional (default: 0.)
        Min value of the uniform distribution.
    upper : float, optional (default: 1.5)
        Max value of the uniform distribution.

    Returns : :class:`numpy.ndarray`
        Attribute value for each edge in `graph`.
    '''
    num = _compute_num_prop(elist, graph, num)
    return np.random.uniform(lower, upper, num)


def gaussian_distrib(graph, elist=None, num=None, avg=None, std=None,
                     **kwargs):
    '''
    Gaussian distribution for edge attributes.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Graph for which an edge attribute will be generated.
    elist : list of edges, optional (default: all edges)
        Generate values for only a subset of edges.
    avg : float, optional (default: 0.)
        Average of the Gaussian distribution.
    std : float, optional (default: 1.5)
        Standard deviation of the Gaussian distribution.

    Returns : :class:`numpy.ndarray`
        Attribute value for each edge in `graph`.
    '''
    num = _compute_num_prop(elist, graph, num)
    return np.random.normal(avg, std, num)


def lognormal_distrib(graph, elist=None, num=None, position=None, scale=None,
                      **kwargs):
    '''
    Lognormal distribution for edge attributes.

    Parameters
    ----------
    graph : :class:`~nngt.Graph` or subclass
        Graph for which an edge attribute will be generated.
    elist : list of edges, optional (default: all edges)
        Generate values for only a subset of edges.
    position : float, optional (default: 0.)
        Average of the normal distribution (i.e. log of the actual mean of the
        lognormal distribution).
    scale : float, optional (default: 1.5)
        Standard deviation of the normal distribution.

    Returns : :class:`numpy.ndarray`
        Attribute value for each edge in `graph`.
    '''
    num = _compute_num_prop(elist, graph, num)
    return np.random.lognormal(position, scale, num)


def lin_correlated_distrib(graph, elist=None, correl_attribute="betweenness",
                           noise_scale=None, lower=None, upper=None,
                           slope=None, offset=0., last_edges=False, **kwargs):
    if slope is not None and (lower, upper) != (None, None):
        raise InvalidArgument('`slope` and `lower`/`upper` parameters are not '
                              'compatible, please choose one or the other.')
    elif (lower is not None or upper is not None) and None in (lower, upper):
        raise InvalidArgument('Both `lower` and `upper` should be set if one '
                              'of the two is used.')
    ecount = _compute_num_prop(elist, graph)
    noise = (1. if noise_scale is None
             else np.abs(np.random.normal(1, noise_scale, ecount)))
    data = None
    if correl_attribute == "betweenness":
        data = graph.get_betweenness(kwargs["btype"], kwargs["weights"])
    elif correl_attribute == "distance":
        assert 'distance' in graph.edge_attributes, \
            'Graph has no "distance" edge attribute.'
        if 'distance' not in kwargs:
            if last_edges:
                slc  = slice(-len(elist), None)
                data = graph.get_edge_attributes(slc, 'distance')
            else:
                data = graph.get_edge_attributes(elist, 'distance')
        else:
            data = kwargs['distance']
    else:
        raise NotImplementedError()
    if noise_scale is not None:
        data *= noise
    if len(data):
        if slope is None:
            dmax = np.max(data)
            dmin = np.min(data)
            return lower + (upper-lower)*(data-dmin)/(dmax-dmin) + offset
        else:
            return slope*data + offset
    return np.array([])


def log_correlated_distrib(graph, elist=None, correl_attribute="betweenness",
                           noise_scale=None, lower=0., upper=2.,
                           **kwargs):
    ecount = _compute_num_prop(elist, graph)
    raise NotImplementedError()


def custom(graph, values=None, elist=None, **kwargs):
    if values is None and elist is not None:
        return np.ones(len(elist))

    return values


di_dfunc = {
    "constant": delta_distrib,
    "uniform": uniform_distrib,
    "lognormal": lognormal_distrib,
    "gaussian": gaussian_distrib,
    "normal": gaussian_distrib,
    "lin_corr": lin_correlated_distrib,
    "log_corr": log_correlated_distrib,
    "custom": custom
}


# ----- #
# Tools #
# ----- #

def _compute_num_prop(elist, graph, ecount=None):
    if ecount is None:
        return len(elist) if elist is not None else graph.edge_nb()
    return ecount
